import pytest
from pr2test.context_manager import make_test_matrix, skip_if_not_supported
from pr2test.marks import require_root
from pr2test.tools import address_exists, interface_exists

from pyroute2 import NetlinkError

pytestmark = [require_root()]

test_matrix = make_test_matrix(
    targets=['local', 'netns'], dbs=['sqlite3/:memory:', 'postgres/pr2test']
)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_context_manager(context):

    ifname = context.new_ifname
    address = '00:11:22:36:47:58'
    spec = {'ifname': ifname, 'kind': 'dummy'}

    ifobj = context.ndb.interfaces.create(**spec)

    with ifobj:
        pass

    assert interface_exists(context.netns, ifname=ifname, state='down')

    with ifobj:
        ifobj['state'] = 'up'
        ifobj['address'] = address

    assert interface_exists(
        context.netns, ifname=ifname, address=address, state='up'
    )

    with ifobj:
        ifobj.remove()

    assert not interface_exists(context.netns, ifname=ifname)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_fail(context):

    ifname = context.new_ifname
    kind = context.new_ifname
    spec = {'ifname': ifname, 'kind': kind}
    ifobj = context.ndb.interfaces.create(**spec)
    save = dict(ifobj)

    try:
        ifobj.commit()
    except NetlinkError as e:
        assert e.code == 95  # Operation not supported

    assert save == dict(ifobj)
    assert ifobj.state == 'invalid'
    assert not interface_exists(context.netns, ifname=ifname)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_veth_simple(context):
    ifname = context.new_ifname
    peername = context.new_ifname
    spec = {'ifname': ifname, 'peer': peername, 'kind': 'veth'}

    context.ndb.interfaces.create(**spec).commit()

    spec_ifl = {'ifname': ifname}
    spec_pl = {'ifname': peername}

    iflink = context.ndb.interfaces[spec_ifl]['link']
    plink = context.ndb.interfaces[spec_pl]['link']

    assert iflink == context.ndb.interfaces[spec_pl]['index']
    assert plink == context.ndb.interfaces[spec_ifl]['index']
    assert interface_exists(context.netns, ifname=ifname)
    assert interface_exists(context.netns, ifname=peername)

    context.ndb.interfaces[spec_ifl].remove().commit()

    assert not interface_exists(context.netns, ifname=ifname)
    assert not interface_exists(context.netns, ifname=peername)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_veth_spec(context):
    ifname = context.new_ifname
    peername = context.new_ifname
    nsname = context.new_nsname

    context.ndb.sources.add(netns=nsname)

    spec = {
        'ifname': ifname,
        'kind': 'veth',
        'peer': {
            'ifname': peername,
            'address': '00:11:22:33:44:55',
            'net_ns_fd': nsname,
        },
    }
    context.ndb.interfaces.create(**spec).commit()
    context.ndb.interfaces.wait(target=nsname, ifname=peername, timeout=5)

    iflink = context.ndb.interfaces[{'ifname': ifname}]['link']
    plink = context.ndb.interfaces[{'target': nsname, 'ifname': peername}][
        'link'
    ]

    assert iflink == (
        context.ndb.interfaces[{'target': nsname, 'ifname': peername}]['index']
    )
    assert plink == (context.ndb.interfaces[{'ifname': ifname}]['index'])

    assert interface_exists(context.netns, ifname=ifname)
    assert interface_exists(nsname, ifname=peername)
    assert not interface_exists(nsname, ifname=ifname)
    assert not interface_exists(context.netns, ifname=peername)

    (context.ndb.interfaces[{'ifname': ifname}].remove().commit())

    assert not interface_exists(context.netns, ifname=ifname)
    assert not interface_exists(nsname, ifname=ifname)
    assert not interface_exists(context.netns, ifname=peername)
    assert not interface_exists(nsname, ifname=peername)

    context.ndb.sources.remove(nsname)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_dummy(context):

    ifname = context.new_ifname
    spec = {'ifname': ifname, 'kind': 'dummy', 'address': '00:11:22:33:44:55'}
    context.ndb.interfaces.create(**spec).commit()
    assert interface_exists(
        context.netns, ifname=ifname, address='00:11:22:33:44:55'
    )


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_bridge(context):

    bridge = context.new_ifname
    brport = context.new_ifname
    spec_br = {'ifname': bridge, 'kind': 'bridge'}
    spec_pt = {'ifname': brport, 'kind': 'dummy'}

    (context.ndb.interfaces.create(**spec_br).commit())

    (
        context.ndb.interfaces.create(**spec_pt)
        .set('master', context.ndb.interfaces[spec_br]['index'])
        .commit()
    )

    assert interface_exists(context.netns, ifname=bridge)
    assert interface_exists(
        context.netns,
        ifname=brport,
        master=context.ndb.interfaces[spec_br]['index'],
    )


@pytest.mark.parametrize('context', test_matrix, indirect=True)
@skip_if_not_supported
def test_vrf(context):
    vrf = context.new_ifname
    spec = {'ifname': vrf, 'kind': 'vrf'}
    (context.ndb.interfaces.create(**spec).set('vrf_table', 42).commit())
    assert interface_exists(context.netns, ifname=vrf)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_vlan(context):
    host = context.new_ifname
    vlan = context.new_ifname
    spec_host = {'ifname': host, 'kind': 'dummy'}
    spec_vlan = {'ifname': vlan, 'kind': 'vlan'}
    (context.ndb.interfaces.create(**spec_host).commit())
    (
        context.ndb.interfaces.create(**spec_vlan)
        .set('link', context.ndb.interfaces[spec_host]['index'])
        .set('vlan_id', 101)
        .commit()
    )
    assert interface_exists(context.netns, ifname=vlan)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_vxlan(context):
    host = context.new_ifname
    vxlan = context.new_ifname
    spec_host = {'ifname': host, 'kind': 'dummy'}
    spec_vxlan = {'ifname': vxlan, 'kind': 'vxlan'}
    (context.ndb.interfaces.create(**spec_host).commit())
    (
        context.ndb.interfaces.create(**spec_vxlan)
        .set('vxlan_link', context.ndb.interfaces[spec_host]['index'])
        .set('vxlan_id', 101)
        .set('vxlan_group', '239.1.1.1')
        .set('vxlan_ttl', 16)
        .commit()
    )
    assert interface_exists(context.netns, ifname=vxlan)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_basic_address(context):

    ifaddr = context.new_ipaddr
    ifname = context.new_ifname
    spec_if = {'ifname': ifname, 'kind': 'dummy', 'state': 'up'}
    i = context.ndb.interfaces.create(**spec_if)
    i.commit()

    spec_ad = {'index': i['index'], 'address': ifaddr, 'prefixlen': 24}
    a = context.ndb.addresses.create(**spec_ad)
    a.commit()
    assert interface_exists(context.netns, ifname=ifname)
    assert address_exists(context.netns, ifname=ifname, address=ifaddr)



from socket import AF_INET, AF_INET6

import pytest
from pr2test.context_manager import make_test_matrix
from pr2test.marks import require_root
from pr2test.tools import address_exists, interface_exists

pytestmark = [require_root()]
test_matrix = make_test_matrix(
    targets=['local', 'netns'], dbs=['sqlite3/:memory:', 'postgres/pr2test']
)


@pytest.mark.parametrize(
    'ipam,prefixlen', (('new_ipaddr', 24), ('new_ip6addr', 64))
)
@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_add_del_ip_dict(context, ipam, prefixlen):
    ifname = context.new_ifname
    ifaddr1 = getattr(context, ipam)
    ifaddr2 = getattr(context, ipam)

    (
        context.ndb.interfaces.create(
            ifname=ifname, kind='dummy', state='down'
        )
        .add_ip({'address': ifaddr1, 'prefixlen': prefixlen})
        .add_ip({'address': ifaddr2, 'prefixlen': prefixlen})
        .commit()
    )

    assert address_exists(context.netns, ifname=ifname, address=ifaddr1)
    assert address_exists(context.netns, ifname=ifname, address=ifaddr2)

    (
        context.ndb.interfaces[{'ifname': ifname}]
        .del_ip({'address': ifaddr2, 'prefixlen': prefixlen})
        .del_ip({'address': ifaddr1, 'prefixlen': prefixlen})
        .commit()
    )

    assert not address_exists(context.netns, ifname=ifname, address=ifaddr1)
    assert not address_exists(context.netns, ifname=ifname, address=ifaddr2)


@pytest.mark.parametrize(
    'ipam,prefixlen', (('new_ipaddr', 24), ('new_ip6addr', 64))
)
@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_add_del_ip_string(context, ipam, prefixlen):
    ifname = context.new_ifname
    ifaddr1 = f'{getattr(context, ipam)}/{prefixlen}'
    ifaddr2 = f'{getattr(context, ipam)}/{prefixlen}'

    (
        context.ndb.interfaces.create(
            ifname=ifname, kind='dummy', state='down'
        )
        .add_ip(ifaddr1)
        .add_ip(ifaddr2)
        .commit()
    )

    assert address_exists(context.netns, ifname=ifname, address=ifaddr1)
    assert address_exists(context.netns, ifname=ifname, address=ifaddr2)

    (
        context.ndb.interfaces[{'ifname': ifname}]
        .del_ip(ifaddr2)
        .del_ip(ifaddr1)
        .commit()
    )

    assert not address_exists(context.netns, ifname=ifname, address=ifaddr1)
    assert not address_exists(context.netns, ifname=ifname, address=ifaddr2)


@pytest.mark.parametrize(
    'ipam,prefixlen,family',
    (('new_ipaddr', 24, AF_INET), ('new_ip6addr', 64, AF_INET6)),
)
@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_del_ip_match(context, ipam, prefixlen, family):
    ifname = context.new_ifname
    ipaddr1 = getattr(context, ipam)
    ipaddr2 = getattr(context, ipam)
    ipaddr3 = getattr(context, ipam)

    (
        context.ndb.interfaces.create(ifname=ifname, kind='dummy', state='up')
        .add_ip(address=ipaddr1, prefixlen=prefixlen)
        .add_ip(address=ipaddr2, prefixlen=prefixlen)
        .add_ip(address=ipaddr3, prefixlen=prefixlen)
        .commit()
    )

    assert address_exists(context.netns, ifname=ifname, address=ipaddr1)
    assert address_exists(context.netns, ifname=ifname, address=ipaddr2)
    assert address_exists(context.netns, ifname=ifname, address=ipaddr3)

    (context.ndb.interfaces[ifname].del_ip(family=family).commit())

    assert not address_exists(context.netns, ifname=ifname, address=ipaddr1)
    assert not address_exists(context.netns, ifname=ifname, address=ipaddr2)
    assert not address_exists(context.netns, ifname=ifname, address=ipaddr3)


@pytest.mark.parametrize(
    'ipam,prefixlen', (('new_ipaddr', 24), ('new_ip6addr', 64))
)
@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_del_ip_fail(context, ipam, prefixlen):
    ifname = context.new_ifname
    ipaddr = f'{getattr(context, ipam)}/{prefixlen}'
    ipaddr_fail = f'{getattr(context, ipam)}/{prefixlen}'

    (
        context.ndb.interfaces.create(ifname=ifname, kind='dummy', state='up')
        .add_ip(ipaddr)
        .commit()
    )

    assert interface_exists(context.netns, ifname=ifname)
    assert address_exists(context.netns, ifname=ifname, address=ipaddr)

    with pytest.raises(KeyError):
        (context.ndb.interfaces[ifname].del_ip(ipaddr_fail).commit())

    assert interface_exists(context.netns, ifname=ifname)
    assert address_exists(context.netns, ifname=ifname, address=ipaddr)


@pytest.mark.parametrize(
    'ipam,prefixlen', (('new_ipaddr', 24), ('new_ip6addr', 64))
)
@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_del_ip_match_fail(context, ipam, prefixlen):
    ifname = context.new_ifname
    ipaddr = getattr(context, ipam)
    ipaddr_fail = getattr(context, ipam)

    (
        context.ndb.interfaces.create(ifname=ifname, kind='dummy', state='up')
        .add_ip(address=ipaddr, prefixlen=prefixlen)
        .commit()
    )

    assert interface_exists(context.netns, ifname=ifname)
    assert address_exists(context.netns, ifname=ifname, address=ipaddr)

    with pytest.raises(KeyError):
        (context.ndb.interfaces[ifname].del_ip(address=ipaddr_fail).commit())

    assert interface_exists(context.netns, ifname=ifname)
    assert address_exists(context.netns, ifname=ifname, address=ipaddr)



import time
from functools import partial

import pytest
from pr2test.context_manager import make_test_matrix
from pr2test.marks import require_root
from pr2test.tools import interface_exists
from utils import require_user

from pyroute2 import config

pytestmark = [require_root()]

test_matrix = make_test_matrix(
    targets=['local', 'netns'], dbs=['sqlite3/:memory:', 'postgres/pr2test']
)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_view_cache(context):
    '''
    NDB stores all the info in an SQL database, and instantiates
    python objects only upon request, since it isn't cheap.
    References to the created objects are stored in the object
    cache until expired.

    This test checks is the cache works as expected. Initially
    there should be no references in the cache, check if the
    references are properly cached and expired in time.
    '''
    require_user('root')
    ifname1 = context.new_ifname
    ifname2 = context.new_ifname

    ndb = context.ndb

    #
    # the cache is empty from the beginning
    assert len(list(ndb.interfaces.cache)) == 0
    #
    # create test interfaces
    ndb.interfaces.create(ifname=ifname1, kind='dummy').commit()
    ndb.interfaces.create(ifname=ifname2, kind='dummy').commit()
    assert interface_exists(context.netns, ifname=ifname1)
    assert interface_exists(context.netns, ifname=ifname2)
    #
    # the interface object must not be cached, as they
    # weren't referenced yet
    assert len(list(ndb.interfaces.cache)) == 0
    #
    # setup the cache expiration time
    ce = config.cache_expire  # save the old value
    config.cache_expire = 1  # set the new one
    #
    # access the interfaces via __getitem__() -- this must
    # create objects and cache the references
    assert ndb.interfaces[ifname1] is not None
    assert ndb.interfaces[ifname2] is not None
    #
    # both references must be in the cache now
    assert len(list(ndb.interfaces.cache)) == 2
    #
    # expire the cache
    time.sleep(1)
    #
    # access the second interface to trigger the
    # cache invalidation
    assert ndb.interfaces[ifname2] is not None
    #
    # ifname1 must be out of the cache now as not
    # accessed within the timeout
    #
    # only ifname2 must remain
    assert len(list(ndb.interfaces.cache)) == 1
    assert list(ndb.interfaces.cache.items())[0][1]['ifname'] == ifname2
    #
    # restore the environment
    config.cache_expire = ce
    ndb.interfaces[ifname1].remove().commit()
    ndb.interfaces[ifname2].remove().commit()

    #
    # check that the interfaces are cleaned up from the system
    assert not interface_exists(context.netns, ifname=ifname1)
    assert not interface_exists(context.netns, ifname=ifname2)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_readonly(context):
    readonly = context.ndb.readonly()

    with pytest.raises(PermissionError):
        readonly.interfaces.create(ifname='test', kind='dummy')

    selection = list(readonly.interfaces.summary().filter(ifname='lo'))
    assert len(selection) == 1
    assert selection[0].ifname == 'lo'
    assert selection[0].address == '00:00:00:00:00:00'
    assert selection[0].target == 'localhost'


@pytest.mark.parametrize('method', ('dump', 'summary'))
@pytest.mark.parametrize(
    'view,sub,func',
    (
        ('routes', 'routes', lambda index, x: x.oif == index),
        ('addresses', 'ipaddr', lambda index, x: x.index == index),
        ('interfaces', 'ports', lambda index, x: x.master == index),
        ('neighbours', 'neighbours', lambda index, x: x.ifindex == index),
        ('vlans', 'vlans', lambda index, x: x.index == index),
    ),
)
@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_nested_count(context, view, sub, func, method):
    br0 = context.new_ifname
    br0p0 = context.new_ifname
    br0p1 = context.new_ifname
    ipaddr1 = context.new_ipaddr
    ipaddr2 = context.new_ipaddr
    gateway = context.new_ipaddr
    net = context.new_ip4net

    context.ndb.interfaces.create(
        ifname=br0p0, kind='dummy', state='up'
    ).commit()
    context.ndb.interfaces.create(
        ifname=br0p1, kind='dummy', state='up'
    ).commit()
    (
        context.ndb.interfaces.create(ifname=br0, kind='bridge', state='up')
        .add_port(br0p0)
        .add_port(br0p1)
        .add_ip(f'{ipaddr1}/24')
        .add_ip(f'{ipaddr2}/24')
        .commit()
    )
    context.ndb.routes.create(
        dst=net.network, dst_len=net.netmask, gateway=gateway
    ).commit()

    records_a = (
        getattr(context.ndb, view)
        .dump()
        .filter(partial(func, context.ndb.interfaces[br0]['index']))
    )
    records_b = getattr(getattr(context.ndb.interfaces[br0], sub), method)()
    count = getattr(context.ndb.interfaces[br0], sub).count()
    assert records_b.count() == records_a.count() == count
    assert count < getattr(context.ndb, view).count() or count == 0



import sqlite3

import pytest

from pyroute2 import NDB

try:
    import psycopg2
except ImportError:
    pytest.skip('no psycopg2 module installed', allow_module_level=True)


def test_no_cleanup(spec):

    # start and stop the DB, leaving all the data in the DB file
    NDB(
        db_provider='sqlite3',
        db_spec=spec.db_spec,
        db_cleanup=False,
        log=spec.log_spec,
    ).close()

    # open the DB file
    db = sqlite3.connect(spec.db_spec)
    cursor = db.cursor()
    cursor.execute('SELECT * FROM interfaces')
    interfaces = cursor.fetchall()

    # at least two records: idx 0 and loopback
    assert len(interfaces) > 1
    # all the interfaces must be of the same source, 'localhost'
    assert set([x[0] for x in interfaces]) == set(('localhost',))


def test_postgres_fail(spec):

    try:
        NDB(
            db_provider='postgres',
            db_spec={'dbname': 'some-nonsense-db-name'},
            log=spec.log_spec,
        ).close()
    except psycopg2.OperationalError:
        return

    raise Exception('postgresql exception was expected')



from pr2test.marks import require_root

pytestmark = [require_root()]


def test_nla_operators(context):
    ifname = context.new_ifname
    ipaddr1 = context.new_ipaddr
    ipaddr2 = context.new_ipaddr
    interface = (
        context.ndb.interfaces.create(ifname=ifname, kind='dummy', state='up')
        .add_ip(f'{ipaddr1}/24')
        .add_ip(f'{ipaddr2}/24')
        .commit()
    )

    r = context.ipr.addr('dump', index=interface['index'])
    complement = r[0] - r[1]
    intersection = r[0] & r[1]

    assert complement.get_attr('IFA_ADDRESS') == ipaddr1
    assert complement.get_attr('IFA_LABEL') is None
    assert complement['prefixlen'] == 0
    assert complement['index'] == 0

    assert intersection.get_attr('IFA_ADDRESS') is None
    assert intersection.get_attr('IFA_LABEL') == ifname
    assert intersection['prefixlen'] == 24
    assert intersection['index'] == context.ndb.interfaces[ifname]['index']


def test_nla_compare(context):
    lvalue = context.ipr.get_links()
    rvalue = context.ipr.get_links()
    assert lvalue is not rvalue
    if lvalue == rvalue:
        pass
    if lvalue != rvalue:
        pass
    assert lvalue != 42



import errno
import functools
import getpass
import itertools
import logging
import os
import sys
import uuid
from collections import namedtuple
from socket import AF_INET, AF_INET6

import pytest
from utils import allocate_network, free_network

from pyroute2 import netns
from pyroute2.common import basestring, uifname
from pyroute2.iproute.linux import IPRoute
from pyroute2.ndb.main import NDB
from pyroute2.netlink.exceptions import NetlinkError
from pyroute2.netlink.generic.wireguard import WireGuard
from pyroute2.nslink.nslink import NetNS


def skip_if_not_implemented(func):
    @functools.wraps(func)
    def test_wrapper(context):
        try:
            return func(context)
        except (AttributeError, NotImplementedError):
            pytest.skip('feature not implemented')

    return test_wrapper


def skip_if_not_supported(func):
    @functools.wraps(func)
    def test_wrapper(*argv, **kwarg):
        try:
            return func(*argv, **kwarg)
        except NetlinkError as e:
            if e.code in {errno.EOPNOTSUPP, errno.ENOENT}:
                pytest.skip('feature not supported by platform')
            raise
        except RuntimeError as e:
            pytest.skip(*e.args)

    return test_wrapper


def make_test_matrix(targets=None, tables=None, dbs=None, types=None):
    targets = targets or ['local']
    tables = tables or [None]
    types = types or [None]
    dbs = dbs or ['sqlite3/:memory:']
    ret = []
    skipdb = list(filter(lambda x: x, os.environ.get('SKIPDB', '').split(':')))
    for db in dbs:
        db_provider, db_spec = db.split('/')
        if any(map(db_provider.startswith, skipdb)):
            continue
        if db_provider != 'sqlite3':
            db_spec = {'dbname': db_spec}
            user = os.environ.get('PGUSER')
            port = os.environ.get('PGPORT')
            host = os.environ.get('PGHOST')
            if user:
                db_spec['user'] = user
            if host:
                if not port:
                    db_spec['port'] = 5432
                db_spec['host'] = host
            if port:
                if not host:
                    db_spec['host'] = 'localhost'
                db_spec['port'] = port

        for target in targets:
            for table in tables:
                for kind in types:
                    param_id = f'db={db} ' f'target={target}'
                    if table is not None:
                        param_id += f' table={table}'
                    if kind is not None:
                        param_id += f' kind={kind}'
                    param = pytest.param(
                        ContextParams(
                            db_provider, db_spec, target, table, kind
                        ),
                        id=param_id,
                    )
                    ret.append(param)
    return ret


ContextParams = namedtuple(
    'ContextParams', ('db_provider', 'db_spec', 'target', 'table', 'kind')
)

Interface = namedtuple('Interface', ('index', 'ifname'))
Network = namedtuple('Network', ('family', 'network', 'netmask'))


class SpecContextManager(object):
    '''
    Prepare simple common variables
    '''

    def __init__(self, request, tmpdir):
        self.uid = str(uuid.uuid4())
        pid = os.getpid()
        self.log_base = f'{tmpdir}/ndb-{pid}'
        self.log_spec = (f'{self.log_base}-{self.uid}.log', logging.DEBUG)
        self.db_spec = f'{self.log_base}-{self.uid}.sql'

    def teardown(self):
        pass


class NDBContextManager(object):
    '''
    This class is used to manage fixture contexts.

    * create log spec
    * create NDB with specified parameters
    * provide methods to register interfaces
    * automatically remove registered interfaces
    '''

    def __init__(self, request, tmpdir, **kwarg):

        self.spec = SpecContextManager(request, tmpdir)
        self.netns = None
        #
        # the cleanup registry
        self.interfaces = {}
        self.namespaces = {}

        if 'log' not in kwarg:
            kwarg['log'] = self.spec.log_spec
        if 'rtnl_debug' not in kwarg:
            kwarg['rtnl_debug'] = True

        target = 'local'
        self.table = None
        self.kind = None
        kwarg['db_provider'] = 'sqlite3'
        kwarg['db_spec'] = ':memory:'
        if hasattr(request, 'param'):
            if isinstance(request.param, ContextParams):
                target = request.param.target
                self.table = request.param.table
                self.kind = request.param.kind
                kwarg['db_provider'] = request.param.db_provider
                kwarg['db_spec'] = request.param.db_spec
            elif isinstance(request.param, (tuple, list)):
                target, self.table = request.param
            else:
                target = request.param

        if target == 'local':
            sources = [{'target': 'localhost', 'kind': 'local'}]
        elif target == 'netns':
            self.netns = self.new_nsname
            sources = [
                {'target': 'localhost', 'kind': 'netns', 'netns': self.netns}
            ]
        else:
            sources = None

        if sources is not None:
            kwarg['sources'] = sources
        #
        # select the DB to work on
        db_name = os.environ.get('PYROUTE2_TEST_DBNAME')
        if isinstance(db_name, basestring) and len(db_name):
            kwarg['db_provider'] = 'psycopg2'
            kwarg['db_spec'] = {'dbname': db_name}
        #
        # this instance is to be tested, so do NOT use it
        # in utility methods
        self.db_provider = kwarg['db_provider']
        self.ndb = NDB(**kwarg)
        self.ipr = self.ndb.sources['localhost'].nl.clone()
        self.wg = WireGuard()
        #
        # IPAM
        self.ipnets = [allocate_network() for _ in range(3)]
        self.ipranges = [[str(x) for x in net] for net in self.ipnets]
        self.ip6net = allocate_network(AF_INET6)
        self.ip6counter = itertools.count(1024)
        self.allocated_networks = {AF_INET: [], AF_INET6: []}
        #
        # RPDB objects for cleanup
        self.rules = []
        #
        # default interface (if running as root)
        if getpass.getuser() == 'root':
            ifname = self.new_ifname
            index = self.ndb.interfaces.create(
                ifname=ifname, kind='dummy', state='up'
            ).commit()['index']
            self.default_interface = Interface(index, ifname)
        else:
            self.default_interface = Interface(1, 'lo')

    def register(self, ifname=None, netns=None):
        '''
        Register an interface in `self.interfaces`. If no interface
        name specified, create a random one.

        All the saved interfaces will be removed on `teardown()`
        '''
        if ifname is None:
            ifname = uifname()
        self.interfaces[ifname] = netns
        return ifname

    def register_netns(self, netns=None):
        '''
        Register netns in `self.namespaces`. If no netns name is
        specified, create a random one.

        All the save namespaces will be removed on `teardown()`
        '''
        if netns is None:
            netns = str(uuid.uuid4())
        self.namespaces[netns] = None
        return netns

    def register_rule(self, spec, netns=None):
        '''
        Register IP rule for cleanup on `teardown()`.
        '''
        self.rules.append((netns, spec))
        return spec

    def register_network(self, family=AF_INET, network=None):
        '''
        Register or allocate a network.

        All the allocated networks should be deallocated on `teardown()`.
        '''
        if network is None:
            network = allocate_network(family)
        # regsiter for cleanup
        self.allocated_networks[family].append(network)
        # return a simple convenient named tuple
        return Network(family, network.network.format(), network.prefixlen)

    def get_ipaddr(self, r=0):
        '''
        Returns an ip address from the specified range.
        '''
        return str(self.ipranges[r].pop())

    def get_ip6addr(self, r=0):
        '''
        Returns an ip6 address from the specified range.
        '''
        return str(self.ip6net[next(self.ip6counter)])

    @property
    def new_log(self, uid=None):
        uid = uid or str(uuid.uuid4())
        return f'{self.spec.log_base}-{uid}.log'

    @property
    def new_ifname(self):
        '''
        Returns a new unique ifname and registers it to be
        cleaned up on `self.teardown()`
        '''
        return self.register()

    @property
    def new_ipaddr(self):
        '''
        Returns a new ipaddr from the configured range
        '''
        return self.get_ipaddr()

    @property
    def new_ip6addr(self):
        '''
        Returns a new ip6addr from the configured range
        '''
        return self.get_ip6addr()

    @property
    def new_ip4net(self):
        '''
        Returns a new IPv4 network
        '''
        return self.register_network(family=AF_INET)

    @property
    def new_ip6net(self):
        '''
        Returns a new IPv6 network
        '''
        return self.register_network(family=AF_INET6)

    @property
    def new_nsname(self):
        '''
        Returns a new unique nsname and registers it to be
        removed on `self.teardown()`
        '''
        return self.register_netns()

    def teardown(self):
        '''
        1. close the test NDB
        2. remove the registered interfaces, ignore not existing
        '''
        # save postmortem DB for SQLite3
        if self.db_provider == 'sqlite3' and sys.version_info >= (3, 7):
            self.ndb.backup(f'{self.spec.uid}-post.db')
        self.ndb.close()
        self.ipr.close()
        self.wg.close()
        for (ifname, nsname) in self.interfaces.items():
            try:
                ipr = None
                #
                # spawn ipr to remove the interface
                if nsname is not None:
                    ipr = NetNS(nsname)
                else:
                    ipr = IPRoute()
                #
                # lookup the interface index
                index = list(ipr.link_lookup(ifname=ifname))
                if len(index):
                    index = index[0]
                else:
                    #
                    # ignore not existing interfaces
                    continue
                #
                # try to remove it
                ipr.link('del', index=index)
            except NetlinkError as e:
                #
                # ignore if removed (t.ex. by another process)
                if e.code != errno.ENODEV:
                    raise
            finally:
                if ipr is not None:
                    ipr.close()
        for nsname in self.namespaces:
            try:
                netns.remove(nsname)
            except FileNotFoundError:
                pass
        for nsname, rule in self.rules:
            try:
                ipr = None
                if nsname is not None:
                    ipr = NetNS(nsname)
                else:
                    ipr = IPRoute()
                ipr.rule('del', **rule)
            except NetlinkError as e:
                if e.code != errno.ENOENT:
                    raise
            finally:
                if ipr is not None:
                    ipr.close()
        for net in self.ipnets:
            free_network(net)
        for family, networks in self.allocated_networks.items():
            for net in networks:
                free_network(net, family)



import os
import subprocess
import sys

import pytest
from pr2test.marks import require_root

from pyroute2 import DQuotSocket

pytestmark = [
    pytest.mark.skipif(
        sys.version_info < (3, 7),
        reason='the test module requires Python > 3.6',
    ),
    require_root(),
]


class DQuotContextManager:
    def __init__(self):
        self.root = 'test_linux/test_dquot'
        self.gz_image = f'{self.root}/dquot.img.gz'
        self.image = f'{self.root}/dquot.img'
        self.loop = None
        self.mnt = f'{self.root}/mnt'
        self.ds = DQuotSocket()

    def setup(self):
        self.run(f'gunzip -c {self.gz_image} >{self.image}', shell=True)
        self.loop = self.run(['losetup', '-f']).stdout.strip().decode('utf-8')
        st_rdev = os.stat(self.loop).st_rdev
        self.major = os.major(st_rdev)
        self.minor = os.minor(st_rdev)
        self.run(['losetup', self.loop, self.image])
        self.run(['mkdir', '-p', self.mnt])
        self.run(['mount', self.loop, self.mnt, '-o', 'quota'])
        self.run(['quotacheck', self.mnt])
        self.run(['quotaon', self.mnt])

    def run(self, cmd, shell=False, check=True, capture_output=True):
        return subprocess.run(
            cmd, shell=shell, check=check, capture_output=capture_output
        )

    def teardown(self):
        self.ds.close()
        self.run(['quotaoff', self.mnt], check=False)
        self.run(['umount', self.mnt], check=False)
        self.run(['losetup', '-d', self.loop], check=False)
        os.unlink(self.image)

    def __enter__(self):
        self.setup()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.teardown()

    def get_one_msg(self):
        for msg in self.ds.get():
            return msg

    def remove_file(self, fname):
        os.unlink(f'{self.mnt}/{fname}')


@pytest.fixture
def mnt():
    with DQuotContextManager() as cm:
        yield cm


def test_basic(mnt):
    mnt.remove_file('test/0')
    msg = mnt.get_one_msg()
    assert msg.get_attr('QUOTA_NL_A_EXCESS_ID') == os.getuid()
    assert msg.get_attr('QUOTA_NL_A_DEV_MAJOR') == mnt.major
    assert msg.get_attr('QUOTA_NL_A_DEV_MINOR') == mnt.minor



from pr2test.marks import require_root

pytestmark = [require_root()]


def test_fdb_vxlan(context):
    ipaddr = context.new_ipaddr
    host_if = context.new_ifname
    vxlan_if = context.new_ifname

    context.ndb.interfaces.create(ifname=host_if, kind='dummy').commit()
    host_idx = context.ndb.interfaces[host_if]['index']
    (
        context.ndb.interfaces.create(
            ifname=vxlan_if, kind='vxlan', vxlan_link=host_idx, vxlan_id=500
        ).commit()
    )
    vxlan_idx = context.ndb.interfaces[vxlan_if]['index']

    # create FDB record
    l2 = '00:11:22:33:44:55'
    (
        context.ipr.fdb(
            'add', lladdr=l2, ifindex=vxlan_idx, vni=600, port=5678, dst=ipaddr
        )
    )
    # dump
    r = context.ipr.fdb('dump', ifindex=vxlan_idx, lladdr=l2)
    assert len(r) == 1
    assert r[0]['ifindex'] == vxlan_idx
    assert r[0].get_attr('NDA_LLADDR') == l2
    assert r[0].get_attr('NDA_DST') == ipaddr
    assert r[0].get_attr('NDA_PORT') == 5678
    assert r[0].get_attr('NDA_VNI') == 600


def test_fdb_bridge_simple(context):

    ifname = context.new_ifname
    (
        context.ndb.interfaces.create(
            ifname=ifname, kind='bridge', state='up'
        ).commit()
    )
    idx = context.ndb.interfaces[ifname]['index']
    # create FDB record
    l2 = '00:11:22:33:44:55'
    context.ipr.fdb('add', lladdr=l2, ifindex=idx)
    # dump FDB
    r = context.ipr.fdb('dump', ifindex=idx, lladdr=l2)
    # one vlan == 1, one w/o vlan
    assert len(r) == 2
    assert len(list(filter(lambda x: x['ifindex'] == idx, r))) == 2
    assert len(list(filter(lambda x: x.get_attr('NDA_VLAN'), r))) == 1
    assert len(list(filter(lambda x: x.get_attr('NDA_MASTER') == idx, r))) == 2
    assert len(list(filter(lambda x: x.get_attr('NDA_LLADDR') == l2, r))) == 2
    r = context.ipr.fdb('dump', ifindex=idx, lladdr=l2, vlan=1)
    assert len(r) == 1
    assert r[0].get_attr('NDA_VLAN') == 1
    assert r[0].get_attr('NDA_MASTER') == idx
    assert r[0].get_attr('NDA_LLADDR') == l2



import errno
import time

import pytest
from pr2test.marks import require_root

from pyroute2 import L2tp, NetlinkError

pytestmark = [require_root()]


@pytest.fixture
def l2ctx(context):
    try:
        context.l2tp = L2tp()
    except NetlinkError as e:
        if e.code == errno.ENOENT:
            pytest.skip('L2TP netlink API not available')
        raise
    context.local_ip = context.get_ipaddr(r=0)  # <- 0.255
    gateway = context.get_ipaddr(r=0)  # <- 0.254 (gateway can not be 0.255)
    context.remote_ip = context.get_ipaddr(r=1)
    context.l2tpeth0 = context.new_ifname
    context.ndb.interfaces[context.default_interface.ifname].add_ip(
        f'{context.local_ip}/24'
    ).commit()
    context.ndb.routes.create(
        dst=str(context.ipnets[1].network), dst_len=24, gateway=gateway
    ).commit()
    yield context
    try:
        context.l2tp.delete_session(tunnel_id=2324, session_id=3435)
    except Exception:
        pass
    try:
        context.l2tp.delete_tunnel(tunnel_id=2324)
    except Exception:
        pass
    context.l2tp.close()


def test_complete(l2ctx):

    # 1. create tunnel
    l2ctx.l2tp.create_tunnel(
        tunnel_id=2324,
        peer_tunnel_id=2425,
        remote=l2ctx.remote_ip,
        local=l2ctx.local_ip,
        udp_dport=32000,
        udp_sport=32000,
        encap="udp",
    )
    tunnel = l2ctx.l2tp.get_tunnel(tunnel_id=2324)
    assert tunnel[0].get_attr("L2TP_ATTR_CONN_ID") == 2324
    assert tunnel[0].get_attr("L2TP_ATTR_PEER_CONN_ID") == 2425
    assert tunnel[0].get_attr("L2TP_ATTR_IP_DADDR") == l2ctx.remote_ip
    assert tunnel[0].get_attr("L2TP_ATTR_IP_SADDR") == l2ctx.local_ip
    assert tunnel[0].get_attr("L2TP_ATTR_UDP_DPORT") == 32000
    assert tunnel[0].get_attr("L2TP_ATTR_UDP_SPORT") == 32000
    assert tunnel[0].get_attr("L2TP_ATTR_ENCAP_TYPE") == 0  # 0 == UDP
    assert tunnel[0].get_attr("L2TP_ATTR_DEBUG") == 0

    # 2. create session
    l2ctx.l2tp.create_session(
        tunnel_id=2324,
        session_id=3435,
        peer_session_id=3536,
        ifname=l2ctx.l2tpeth0,
    )
    session = l2ctx.l2tp.get_session(tunnel_id=2324, session_id=3435)
    assert session[0].get_attr("L2TP_ATTR_SESSION_ID") == 3435
    assert session[0].get_attr("L2TP_ATTR_PEER_SESSION_ID") == 3536
    assert session[0].get_attr("L2TP_ATTR_DEBUG") == 0

    # setting up DEBUG -> 95, operation not supported; review the test
    # # 3. modify session
    # l2ctx.l2tp.modify_session(tunnel_id=2324, session_id=3435, debug=True)
    # session = l2ctx.l2tp.get_session(tunnel_id=2324, session_id=3435)
    # assert session[0].get_attr("L2TP_ATTR_DEBUG") == 1

    # setting up DEBUG -> 95, operation not supported; review the test
    # # 4. modify tunnel
    # l2ctx.l2tp.modify_tunnel(tunnel_id=2324, debug=True)
    # tunnel = l2ctx.l2tp.get_tunnel(tunnel_id=2324)
    # assert tunnel[0].get_attr("L2TP_ATTR_DEBUG") == 1

    # 5. destroy session
    l2ctx.l2tp.delete_session(tunnel_id=2324, session_id=3435)
    for _ in range(5):
        try:
            assert not l2ctx.l2tp.get_session(tunnel_id=2324, session_id=3435)
        except AssertionError:
            time.sleep(0.1)
            continue
        break
    else:
        raise Exception('could not remove L2TP session')

    # 6. destroy tunnel
    l2ctx.l2tp.delete_tunnel(tunnel_id=2324)
    for _ in range(5):
        try:
            assert not l2ctx.l2tp.get_tunnel(tunnel_id=2324)
        except AssertionError:
            time.sleep(0.1)
            continue
        break
    else:
        raise Exception('could not remove L2TP tunnel')



import pytest
from pr2test.marks import require_root

from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg

pytestmark = [require_root()]


@pytest.mark.parametrize('smode', ('IPVLAN_MODE_L2', 'IPVLAN_MODE_L3'))
def test_create_ipvlan(context, smode):
    master = context.new_ifname
    ipvlan = context.new_ifname
    # create the master link
    index = context.ndb.interfaces.create(
        ifname=master, kind='dummy'
    ).commit()['index']
    # check modes
    # maybe move modes dict somewhere else?
    cmode = ifinfmsg.ifinfo.data_map['ipvlan'].modes[smode]
    assert ifinfmsg.ifinfo.data_map['ipvlan'].modes[cmode] == smode
    # create ipvlan
    context.ipr.link(
        'add', ifname=ipvlan, kind='ipvlan', link=index, mode=cmode
    )
    interface = context.ndb.interfaces.wait(ifname=ipvlan, timeout=5)
    assert interface['link'] == index
    assert interface['ipvlan_mode'] == cmode



import pytest
from pr2test.context_manager import make_test_matrix
from pr2test.marks import require_root
from pr2test.tools import interface_exists

pytestmark = [require_root()]

tnl_matrix = make_test_matrix(
    targets=['local', 'netns'],
    types=['gre', 'ipip', 'sit'],
    dbs=['sqlite3/:memory:', 'postgres/pr2test'],
)


def _test_tunnel_endpoints(context, state):
    ifname = context.new_ifname
    ipaddr_local1 = context.new_ipaddr
    ipaddr_local2 = context.new_ipaddr
    ipaddr_remote = context.new_ipaddr
    kind = context.kind

    (
        context.ndb.interfaces.create(
            **{
                'ifname': ifname,
                'state': state,
                'kind': kind,
                f'{kind}_local': ipaddr_local1,
                f'{kind}_remote': ipaddr_remote,
            }
        ).commit()
    )

    def match(ifname, ipaddr):
        return (
            lambda x: x.get_nested('IFLA_LINKINFO', 'IFLA_INFO_KIND') == kind
            and x.get_attr('IFLA_IFNAME') == ifname
            and x.get_nested(
                'IFLA_LINKINFO',
                'IFLA_INFO_DATA',
                'IFLA_%s_LOCAL' % kind.upper(),
            )
            == ipaddr
        )

    assert interface_exists(context.netns, match(ifname, ipaddr_local1))

    (
        context.ndb.interfaces[ifname]
        .set(f'{kind}_local', ipaddr_local2)
        .commit()
    )

    assert interface_exists(context.netns, match(ifname, ipaddr_local2))


@pytest.mark.parametrize('context', tnl_matrix, indirect=True)
def test_tunnel_endpoints_down(context):
    return _test_tunnel_endpoints(context, 'down')


@pytest.mark.parametrize('context', tnl_matrix, indirect=True)
def test_tunnel_endpoints_up(context):
    return _test_tunnel_endpoints(context, 'up')



import os
import pathlib
import sys

import pytest
from pr2test.marks import require_root

pytestmark = [require_root()]


def get_examples(*argv):
    root = pathlib.Path(os.environ['WORKSPACE'])
    examples = [
        file
        for file in root.joinpath(*argv).iterdir()
        if not file.name.endswith('.swp')
    ]
    return {
        'argnames': 'example',
        'argvalues': examples,
        'ids': [x.name for x in examples],
    }


@pytest.mark.parametrize(**get_examples('examples', 'pyroute2-cli'))
def test_cli_examples(example, pytester, context):
    with example.open('r') as text:
        result = pytester.run('pyroute2-cli', stdin=text)
    assert result.ret == 0


@pytest.mark.parametrize(**get_examples('examples', 'ndb'))
def test_ndb_examples(example, pytester, context):
    argv = []
    with example.open('r') as text:
        for line in text.readlines():
            line = line.strip()
            if line == ':notest:':
                pytest.skip()
            elif line.startswith(':test:argv:'):
                argv.append(line.split(':')[-1])
            elif line.startswith(':test:environ:'):
                key, value = line.split(':')[-1].split('=')
                os.environ[key] = value
    result = pytester.run(sys.executable, example.as_posix(), *argv)
    assert result.ret == 0


def test_basic(tmpdir, pytester, context):
    pytester.makefile('.pr2', test='interfaces lo mtu')
    with open('test.pr2', 'r') as text:
        result = pytester.run("pyroute2-cli", stdin=text)
    assert result.ret == 0
    assert result.outlines == ['65536']



import select

import pytest

# NetlinkError, NetlinkDecodeError
# IPRoute
# IPRSocket
from pyroute2 import IPRoute as IPRoute0
from pyroute2 import IPRSocket as IPRSocket0
from pyroute2 import NetlinkDecodeError as NetlinkDecodeError0
from pyroute2 import NetlinkError as NetlinkError0
from pyroute2.iproute import IPRoute as IPRoute1

# nlmsg
# flags
from pyroute2.netlink import (
    NLM_F_DUMP,
    NLM_F_MATCH,
    NLM_F_REQUEST,
    NLM_F_ROOT,
    NLMSG_DONE,
    NLMSG_ERROR,
)
from pyroute2.netlink import NetlinkDecodeError as NetlinkDecodeError1
from pyroute2.netlink import NetlinkError as NetlinkError1
from pyroute2.netlink import nlmsg
from pyroute2.netlink.rtnl import (
    RTM_DELADDR,
    RTM_DELLINK,
    RTM_GETADDR,
    RTM_GETLINK,
    RTM_NEWADDR,
    RTM_NEWLINK,
    RTMGRP_IPV4_IFADDR,
    RTMGRP_IPV6_IFADDR,
    RTMGRP_LINK,
)
from pyroute2.netlink.rtnl.ifaddrmsg import ifaddrmsg
from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
from pyroute2.netlink.rtnl.iprsocket import IPRSocket as IPRSocket1
from pyroute2.netlink.rtnl.rtmsg import rtmsg


def test_exceptions_compat():

    with pytest.raises(NetlinkError1):
        raise NetlinkError1(code=99)

    with pytest.raises(NetlinkDecodeError1):
        raise NetlinkDecodeError1(exception=Exception())


def test_exceptions():

    with pytest.raises(NetlinkError0):
        raise NetlinkError0(code=99)

    with pytest.raises(NetlinkDecodeError0):
        raise NetlinkDecodeError0(exception=Exception())


def test_constants():
    assert issubclass(ifinfmsg, nlmsg)
    assert NLM_F_REQUEST == 1
    assert NLM_F_ROOT == 0x100
    assert NLM_F_MATCH == 0x200
    assert NLM_F_DUMP == (NLM_F_ROOT | NLM_F_MATCH)
    assert NLMSG_DONE == 0x3
    assert NLMSG_ERROR == 0x2
    assert RTM_NEWLINK == 0x10
    assert RTM_DELLINK == 0x11
    assert RTM_GETLINK == 0x12
    assert RTM_NEWADDR == 0x14
    assert RTM_DELADDR == 0x15
    assert RTM_GETADDR == 0x16
    assert RTMGRP_LINK == 0x1
    assert RTMGRP_IPV4_IFADDR == 0x10
    assert RTMGRP_IPV6_IFADDR == 0x100


@pytest.mark.parametrize('socket_class', (IPRSocket0, IPRSocket1))
def test_basic(socket_class):

    ip = socket_class()
    ip.bind()

    # check the `socket` interface compliance
    poll = select.poll()
    poll.register(ip, select.POLLIN | select.POLLPRI)
    poll.unregister(ip)
    ip.close()


@pytest.mark.parametrize('iproute_class', (IPRoute0, IPRoute1))
def test_iproute_message_classes(iproute_class):
    with iproute_class() as ip:
        assert {ifaddrmsg, ifinfmsg, rtmsg} < {type(x) for x in ip.dump()}


@pytest.mark.parametrize('iproute_class', (IPRoute0, IPRoute1))
def test_iproute_message_subclass(iproute_class):
    with iproute_class() as ip:
        assert all([issubclass(type(x), nlmsg) for x in ip.dump()])



import pytest
from pr2test.context_manager import make_test_matrix, skip_if_not_supported
from pr2test.marks import require_root

from pyroute2 import protocols

pytestmark = [require_root()]
test_matrix = make_test_matrix(targets=['local', 'netns'])


def find_action(context, prio=1, filter_name="u32"):
    """Returns the first action with priority `prio`
    under the filter with `filter_name`
    """
    # Fetch filters
    filts = context.ipr.get_filters(index=context.default_interface.index)
    # Find our action
    for i in filts:
        try:
            act = (
                i.get_attr('TCA_OPTIONS')
                .get_attr('TCA_%s_ACT' % filter_name.upper())
                .get_attr('TCA_ACT_PRIO_%d' % prio)
            )
            assert act
            return act
        except AttributeError:
            continue
    raise FileNotFoundError('Action not found')


@pytest.mark.parametrize('context', test_matrix, indirect=True)
@skip_if_not_supported
def test_mirred(context):
    index, ifname = context.default_interface
    # add a htb root
    context.ipr.tc('add', 'htb', index=index, handle='1:', default='20:0')
    # mirred action
    actions = [
        dict(
            kind='mirred', direction='egress', action='mirror', ifindex=index
        ),
        dict(
            kind='mirred', direction='egress', action='redirect', ifindex=index
        ),
    ]
    # create a filter with this action
    context.ipr.tc(
        'add-filter',
        'u32',
        index=index,
        handle='0:0',
        parent='1:0',
        prio=10,
        protocol=protocols.ETH_P_IP,
        target='1:10',
        keys=['0x0006/0x00ff+8'],
        action=actions,
    )

    # Check that we have two mirred actions with the right parameters
    act = find_action(context, 1)
    assert act.get_attr('TCA_ACT_KIND') == 'mirred'
    parms = act.get_attr('TCA_ACT_OPTIONS').get_attr('TCA_MIRRED_PARMS')
    assert parms['eaction'] == 2  # egress mirror, see act_mirred.py
    assert parms['ifindex'] == index
    assert parms['action'] == 3  # TC_ACT_PIPE because action == mirror

    act = find_action(context, 2)
    assert act.get_attr('TCA_ACT_KIND') == 'mirred'
    parms = act.get_attr('TCA_ACT_OPTIONS').get_attr('TCA_MIRRED_PARMS')
    assert parms['eaction'] == 1  # egress redirect, see act_mirred.py
    assert parms['ifindex'] == index
    assert parms['action'] == 4  # TC_ACT_STOLEN because action == redirect


@pytest.mark.parametrize('context', test_matrix, indirect=True)
@skip_if_not_supported
def test_connmark(context):
    index, ifname = context.default_interface
    # add a htb root
    context.ipr.tc('add', 'htb', index=index, handle='1:', default='20:0')
    # connmark action
    action = {'kind': 'connmark', 'zone': 63}
    # create a filter with this action
    context.ipr.tc(
        'add-filter',
        'u32',
        index=index,
        handle='0:0',
        parent='1:0',
        prio=10,
        protocol=protocols.ETH_P_IP,
        target='1:10',
        keys=['0x0006/0x00ff+8'],
        action=action,
    )

    act = find_action(context)

    # Check that it is a connmark action with the right parameters
    assert act.get_attr('TCA_ACT_KIND') == 'connmark'
    parms = act.get_attr('TCA_ACT_OPTIONS').get_attr('TCA_CONNMARK_PARMS')
    assert parms['zone'] == 63



from pr2test.context_manager import skip_if_not_supported
from pr2test.marks import require_root

from pyroute2 import MPTCP

pytestmark = [require_root()]


def get_endpoints(mptcp):
    return dict(
        (
            x.get_nested('MPTCP_PM_ATTR_ADDR', 'MPTCP_PM_ADDR_ATTR_ADDR4'),
            x.get_nested('MPTCP_PM_ATTR_ADDR', 'MPTCP_PM_ADDR_ATTR_ID'),
        )
        for x in mptcp.endpoint('show')
    )


def get_limits(mptcp):
    return [
        (
            x.get_attr('MPTCP_PM_ATTR_SUBFLOWS'),
            x.get_attr('MPTCP_PM_ATTR_RCV_ADD_ADDRS'),
        )
        for x in mptcp.limits('show')
    ][0]


@skip_if_not_supported
def test_enpoint_add_addr4(context):
    with MPTCP() as mptcp:
        ipaddrs = [context.new_ipaddr for _ in range(3)]
        for ipaddr in ipaddrs:
            mptcp.endpoint('add', addr=ipaddr)
        mapping = get_endpoints(mptcp)
        assert set(mapping) >= set(ipaddrs)
        for ipaddr in ipaddrs:
            mptcp.endpoint('del', addr=ipaddr, id=mapping[ipaddr])
        assert not set(get_endpoints(mptcp)).intersection(set(ipaddrs))


@skip_if_not_supported
def test_limits(context):
    with MPTCP() as mptcp:
        save_subflows, save_rcv_add = get_limits(mptcp)
        mptcp.limits('set', subflows=2, rcv_add_addrs=3)
        assert get_limits(mptcp) == (2, 3)
        mptcp.limits('set', subflows=save_subflows, rcv_add_addrs=save_rcv_add)
        assert get_limits(mptcp) == (save_subflows, save_rcv_add)



import pytest
from pr2test.context_manager import make_test_matrix, skip_if_not_supported
from pr2test.marks import require_root
from pr2test.tools import qdisc_exists

from pyroute2 import protocols

pytestmark = [require_root()]
test_matrix = make_test_matrix(targets=['local', 'netns'])


@pytest.mark.parametrize('context', test_matrix, indirect=True)
@skip_if_not_supported
def test_htb(context):
    index, ifname = context.default_interface
    # 8<-----------------------------------------------------
    # root queue, '1:0' handle notation
    context.ipr.tc('add', 'htb', index=index, handle='1:', default='20:0')

    assert qdisc_exists(context.netns, 'htb', ifname=ifname)

    # 8<-----------------------------------------------------
    # classes, both string and int handle notation
    context.ipr.tc(
        'add-class',
        'htb',
        index=index,
        handle='1:1',
        parent='1:0',
        rate='256kbit',
        burst=1024 * 6,
    )
    context.ipr.tc(
        'add-class',
        'htb',
        index=index,
        handle=0x10010,
        parent=0x10001,
        rate='192kbit',
        burst=1024 * 6,
        prio=1,
    )
    context.ipr.tc(
        'add-class',
        'htb',
        index=index,
        handle='1:20',
        parent='1:1',
        rate='128kbit',
        burst=1024 * 6,
        prio=2,
    )
    cls = context.ipr.get_classes(index=index)
    assert len(cls) == 3

    # 8<-----------------------------------------------------
    # leaves, both string and int handle notation
    context.ipr.tc(
        'add', 'sfq', index=index, handle='10:', parent='1:10', perturb=10
    )
    context.ipr.tc(
        'add', 'sfq', index=index, handle=0x200000, parent=0x10020, perturb=10
    )
    qds = [x for x in context.ipr.get_qdiscs() if x['index'] == index]
    types = set([x.get_attr('TCA_KIND') for x in qds])
    assert types == set(('htb', 'sfq'))

    # 8<-----------------------------------------------------
    # filters, both string and int handle notation
    #
    # Please note, that u32 filter requires ethernet protocol
    # numbers, as defined in protocols module. Do not provide
    # here socket.AF_INET and so on.
    #
    context.ipr.tc(
        'add-filter',
        'u32',
        index=index,
        handle='0:0',
        parent='1:0',
        prio=10,
        protocol=protocols.ETH_P_IP,
        target='1:10',
        keys=['0x0006/0x00ff+8', '0x0000/0xffc0+2'],
    )
    context.ipr.tc(
        'add-filter',
        'u32',
        index=index,
        handle=0,
        parent=0x10000,
        prio=10,
        protocol=protocols.ETH_P_IP,
        target=0x10020,
        keys=['0x5/0xf+0', '0x10/0xff+33'],
    )
    # 2 filters + 2 autogenerated
    fls = context.ipr.get_filters(index=index)
    assert len(fls) == 4


@pytest.mark.parametrize('context', test_matrix, indirect=True)
@skip_if_not_supported
def test_replace(context):
    test_htb(context)
    index, ifname = context.default_interface
    # change class
    context.ipr.tc(
        'replace-class',
        'htb',
        index=index,
        handle=0x10010,
        parent=0x10001,
        rate='102kbit',
        burst=1024 * 6,
        prio=3,
    )
    clss = context.ipr.get_classes(index=index)
    for cls in clss:
        if cls['handle'] == 0x10010:
            break
    else:
        raise Exception('target class not found')
    opts = cls.get_attr('TCA_OPTIONS')
    params = opts.get_attr('TCA_HTB_PARMS')

    assert params['prio'] == 3
    assert params['quantum'] * 8 == 10200



from pr2test.marks import require_root

pytestmark = [require_root()]


def callback(msg, cb_context):
    cb_context['counter'] += 1


def test_callbacks_positive(context):
    ifname = context.new_ifname
    cb_context = {'counter': 0}
    interface = context.ndb.interfaces.create(
        ifname=ifname, kind='dummy'
    ).commit()

    context.ipr.register_callback(
        callback,
        lambda x: x.get('index', None) == interface['index'],
        (cb_context,),
    )
    context.ipr.link('set', index=interface['index'], state='up')
    context.ipr.link('get', index=interface['index'])
    counter = cb_context['counter']
    assert counter > 0
    context.ipr.unregister_callback(callback)
    context.ipr.link('set', index=interface['index'], state='down')
    context.ipr.link('get', index=interface['index'])
    assert cb_context['counter'] == counter


def test_callbacks_negative(context):
    ifname = context.new_ifname
    cb_context = {'counter': 0}
    interface = context.ndb.interfaces.create(
        ifname=ifname, kind='dummy'
    ).commit()

    context.ipr.register_callback(
        callback, lambda x: x.get('index', None) == -1, (cb_context,)
    )
    context.ipr.link('set', index=interface['index'], state='up')
    context.ipr.link('get', index=interface['index'])
    counter = cb_context['counter']
    assert counter == 0



from socket import AF_INET, AF_INET6

from pr2test.marks import require_root

pytestmark = [require_root()]


def test_peer_ipv4(context):

    ifname = context.new_ifname
    ipaddr = context.new_ipaddr
    port = 9999
    listen = 2525
    peer_ip_1 = context.new_ipaddr
    peer_ip_2 = context.new_ipaddr
    allowed_ip_1 = str(context.ipnets[1])
    allowed_ip_2 = str(context.ipnets[2])

    (
        context.ndb.interfaces.create(ifname=ifname, kind='wireguard')
        .add_ip(f'{ipaddr}/24')
        .set('state', 'up')
        .commit()
    )

    peer_1 = {
        'public_key': 'TGFHcm9zc2VCaWNoZV9DJ2VzdExhUGx1c0JlbGxlPDM=',
        'endpoint_addr': peer_ip_1,
        'endpoint_port': port,
        'persistent_keepalive': 15,
        'allowed_ips': [f'{allowed_ip_1}'],
    }

    peer_2 = {
        'public_key': 'AGFHcm9zc2VCaWNoZV9DJ2VzdExhUGx1c0JlbGxlPDM=',
        'endpoint_addr': peer_ip_2,
        'endpoint_port': port,
        'persistent_keepalive': 15,
        'allowed_ips': [f'{allowed_ip_2}'],
    }

    (
        context.wg.set(
            ifname,
            private_key='RCdhcHJlc0JpY2hlLEplU2VyYWlzTGFQbHVzQm9ubmU=',
            fwmark=0x1337,
            listen_port=listen,
            peer=peer_1,
        )
    )

    (
        context.wg.set(
            ifname,
            private_key='RCdhcHJlc0JpY2hlLEplU2VyYWlzTGFQbHVzQm9ubmU=',
            fwmark=0x1337,
            listen_port=listen,
            peer=peer_2,
        )
    )

    for peer in context.wg.info(ifname)[0].get_attr('WGDEVICE_A_PEERS'):
        endpoint = peer.get_attr('WGPEER_A_ENDPOINT')
        allowed = peer.get_attr('WGPEER_A_ALLOWEDIPS')
        assert endpoint['family'] == AF_INET
        assert endpoint['port'] == port
        assert endpoint['addr'] in (peer_ip_1, peer_ip_2)
        assert allowed[0]['addr'] in (allowed_ip_1, allowed_ip_2)


def test_peer_ipv6(context):

    ifname = context.new_ifname
    ipaddr = context.new_ipaddr
    port = 9999
    listen = 2525
    peer_ip_1 = '::fa'
    peer_ip_2 = '::fb'
    allowed_ip_1 = 'fa::/64'
    allowed_ip_2 = 'fb::/64'

    (
        context.ndb.interfaces.create(ifname=ifname, kind='wireguard')
        .add_ip(f'{ipaddr}/24')
        .set('state', 'up')
        .commit()
    )

    peer_1 = {
        'public_key': 'TGFHcm9zc2VCaWNoZV9DJ2VzdExhUGx1c0JlbGxlPDM=',
        'endpoint_addr': peer_ip_1,
        'endpoint_port': port,
        'persistent_keepalive': 15,
        'allowed_ips': [f'{allowed_ip_1}'],
    }

    peer_2 = {
        'public_key': 'AGFHcm9zc2VCaWNoZV9DJ2VzdExhUGx1c0JlbGxlPDM=',
        'endpoint_addr': peer_ip_2,
        'endpoint_port': port,
        'persistent_keepalive': 15,
        'allowed_ips': [f'{allowed_ip_2}'],
    }

    (
        context.wg.set(
            ifname,
            private_key='RCdhcHJlc0JpY2hlLEplU2VyYWlzTGFQbHVzQm9ubmU=',
            fwmark=0x1337,
            listen_port=listen,
            peer=peer_1,
        )
    )

    (
        context.wg.set(
            ifname,
            private_key='RCdhcHJlc0JpY2hlLEplU2VyYWlzTGFQbHVzQm9ubmU=',
            fwmark=0x1337,
            listen_port=listen,
            peer=peer_2,
        )
    )

    for peer in context.wg.info(ifname)[0].get_attr('WGDEVICE_A_PEERS'):
        endpoint = peer.get_attr('WGPEER_A_ENDPOINT')
        allowed = peer.get_attr('WGPEER_A_ALLOWEDIPS')
        assert endpoint['family'] == AF_INET6
        assert endpoint['port'] == port
        assert endpoint['addr'] in (peer_ip_1, peer_ip_2)
        assert allowed[0]['addr'] in (allowed_ip_1, allowed_ip_2)



import pytest
from pr2test.context_manager import make_test_matrix
from pr2test.marks import require_root
from pr2test.tools import fdb_record_exists

pytestmark = [require_root()]

test_matrix = make_test_matrix(
    targets=['local', 'netns'],
    tables=[None],
    dbs=['sqlite3/:memory:', 'postgres/pr2test'],
)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_fdb_create(context):

    spec = {
        'ifindex': context.default_interface.index,
        'lladdr': '00:11:22:33:44:55',
    }

    context.ndb.fdb.create(**spec).commit()
    assert fdb_record_exists(context.netns, **spec)

    context.ndb.fdb[spec].remove().commit()
    assert not fdb_record_exists(context.netns, **spec)



from pr2test.marks import require_root

from pyroute2 import NetNS

pytestmark = [require_root()]


def test_get_netns_info(context):
    nsname = context.new_nsname
    peer_name = context.new_ifname
    host_name = context.new_ifname
    with NetNS(nsname):
        (
            context.ndb.interfaces.create(
                ifname=host_name,
                kind='veth',
                peer={'ifname': peer_name, 'net_ns_fd': nsname},
            ).commit()
        )
        # get veth
        veth = context.ipr.link('get', ifname=host_name)[0]
        target = veth.get_attr('IFLA_LINK_NETNSID')
        for info in context.ipr.get_netns_info():
            path = info.get_attr('NSINFO_PATH')
            assert path.endswith(nsname)
            netnsid = info['netnsid']
            if target == netnsid:
                break
        else:
            raise KeyError('peer netns not found')



import os
import socket

from pr2test.marks import require_root

from pyroute2 import NetlinkDumpInterrupted

pytestmark = [require_root()]


def test_mass_ipv6(context):
    #
    ipv6net = context.new_ip6net
    base = str(ipv6net.network) + '{0}'
    limit = int(os.environ.get('PYROUTE2_SLIMIT', '0x800'), 16)
    index, ifname = context.default_interface

    # add addresses
    for idx in range(limit):
        context.ipr.addr(
            'add',
            index=index,
            family=socket.AF_INET6,
            address=base.format(hex(idx)[2:]),
            prefixlen=48,
        )

    # assert addresses in two steps, to ease debug
    addrs = []
    for _ in range(3):
        try:
            addrs = context.ipr.get_addr(family=socket.AF_INET6)
            break
        except NetlinkDumpInterrupted:
            pass
    else:
        raise Exception('could not dump addresses')
    assert len(addrs) >= limit



import pytest
from pr2test.context_manager import make_test_matrix
from pr2test.marks import require_root
from pr2test.tools import address_exists, interface_exists, route_exists

pytestmark = [require_root()]

test_matrix = make_test_matrix(
    targets=['local', 'netns'], dbs=['sqlite3/:memory:', 'postgres/pr2test']
)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_create(context):
    ifname = context.new_ifname
    iface = context.ndb.interfaces.create(ifname=ifname, kind='dummy').commit()
    assert interface_exists(context.netns, ifname=ifname)
    iface.rollback()
    assert not interface_exists(context.netns, ifname=ifname)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_remove(context):
    ifname = context.new_ifname
    iface = context.ndb.interfaces.create(ifname=ifname, kind='dummy').commit()
    assert interface_exists(context.netns, ifname=ifname)
    iface.remove().commit()
    assert not interface_exists(context.netns, ifname=ifname)
    iface.rollback()
    assert interface_exists(context.netns, ifname=ifname)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_set(context):
    ifname = context.new_ifname
    (
        context.ndb.interfaces.create(
            ifname=ifname, kind='dummy', address='00:11:22:33:44:55'
        ).commit()
    )
    assert interface_exists(
        context.netns, ifname=ifname, address='00:11:22:33:44:55'
    )
    (
        context.ndb.interfaces[ifname]
        .set('address', '00:11:22:aa:aa:aa')
        .commit()
    )
    assert not interface_exists(
        context.netns, ifname=ifname, address='00:11:22:33:44:55'
    )
    assert interface_exists(
        context.netns, ifname=ifname, address='00:11:22:aa:aa:aa'
    )
    (context.ndb.interfaces[ifname].rollback())
    assert not interface_exists(
        context.netns, ifname=ifname, address='00:11:22:aa:aa:aa'
    )
    assert interface_exists(
        context.netns, ifname=ifname, address='00:11:22:33:44:55'
    )


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_simple_deps(context):

    ifname = context.new_ifname
    ipaddr = context.new_ipaddr
    router = context.new_ipaddr
    dst = str(context.ipnets[1].network)

    #
    # simple dummy interface with one address and
    # one dependent route
    #
    (
        context.ndb.interfaces.create(ifname=ifname, kind='dummy')
        .set('state', 'up')
        .add_ip(address=ipaddr, prefixlen=24)
        .commit()
    )

    (context.ndb.routes.create(dst=dst, dst_len=24, gateway=router).commit())

    # check everything is in place
    assert interface_exists(context.netns, ifname=ifname)
    assert address_exists(context.netns, ifname=ifname, address=ipaddr)
    assert route_exists(context.netns, gateway=router, dst=dst, dst_len=24)

    # remove the interface
    iface = context.ndb.interfaces[ifname].remove().commit()

    # check there is no interface, no route
    assert not interface_exists(context.netns, ifname=ifname)
    assert not address_exists(context.netns, ifname=ifname, address=ipaddr)
    assert not route_exists(context.netns, gateway=router, dst=dst, dst_len=24)

    # revert the changes using the implicit last_save
    iface.rollback()

    assert interface_exists(context.netns, ifname=ifname)
    assert address_exists(context.netns, ifname=ifname, address=ipaddr)
    assert route_exists(context.netns, gateway=router, dst=dst, dst_len=24)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_bridge_deps(context):

    if_br0 = context.new_ifname
    if_br0p0 = context.new_ifname
    if_br0p1 = context.new_ifname
    ifaddr1 = context.new_ipaddr
    ifaddr2 = context.new_ipaddr
    router = context.new_ipaddr
    dst = str(context.ipnets[1].network)

    with context.ndb.interfaces as i:
        i.create(ifname=if_br0p0, kind='dummy', state='up').commit()
        i.create(ifname=if_br0p1, kind='dummy', state='up').commit()
        (
            i.create(ifname=if_br0, kind='bridge', state='up')
            .add_port(if_br0p0)
            .add_port(if_br0p1)
            .add_ip(address=ifaddr1, prefixlen=24)
            .add_ip(address=ifaddr2, prefixlen=24)
            .commit()
        )

    (context.ndb.routes.create(dst=dst, dst_len=24, gateway=router).commit())

    assert interface_exists(context.netns, ifname=if_br0)
    assert interface_exists(context.netns, ifname=if_br0p0)
    assert interface_exists(context.netns, ifname=if_br0p1)
    assert address_exists(context.netns, ifname=if_br0, address=ifaddr1)
    assert address_exists(context.netns, ifname=if_br0, address=ifaddr2)
    assert route_exists(context.netns, gateway=router, dst=dst, dst_len=24)

    # remove the interface
    iface = context.ndb.interfaces[if_br0].remove().commit()

    assert not interface_exists(context.netns, ifname=if_br0)
    assert not address_exists(context.netns, ifname=if_br0, address=ifaddr1)
    assert not address_exists(context.netns, ifname=if_br0, address=ifaddr2)
    assert not route_exists(context.netns, gateway=router, dst=dst, dst_len=24)

    # revert the changes using the implicit last_save
    iface.rollback()
    assert interface_exists(context.netns, ifname=if_br0)
    assert interface_exists(context.netns, ifname=if_br0p0)
    assert interface_exists(context.netns, ifname=if_br0p1)
    assert address_exists(context.netns, ifname=if_br0, address=ifaddr1)
    assert address_exists(context.netns, ifname=if_br0, address=ifaddr2)
    assert route_exists(context.netns, gateway=router, dst=dst, dst_len=24)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_vlan_deps(context):

    if_host = context.new_ifname
    if_vlan = context.new_ifname
    ifaddr1 = context.new_ipaddr
    ifaddr2 = context.new_ipaddr
    router = context.new_ipaddr
    dst = str(context.ipnets[1].network)

    (
        context.ndb.interfaces.create(
            ifname=if_host, kind='dummy', state='up'
        ).commit()
    )

    (
        context.ndb.interfaces.create(
            ifname=if_vlan, kind='vlan', state='up', vlan_id=1001, link=if_host
        )
        .add_ip(address=ifaddr1, prefixlen=24)
        .add_ip(address=ifaddr2, prefixlen=24)
        .commit()
    )

    (context.ndb.routes.create(dst=dst, dst_len=24, gateway=router).commit())

    # check everything is in place
    assert interface_exists(context.netns, ifname=if_host)
    assert interface_exists(context.netns, ifname=if_vlan)
    assert address_exists(context.netns, ifname=if_vlan, address=ifaddr1)
    assert address_exists(context.netns, ifname=if_vlan, address=ifaddr2)
    assert route_exists(context.netns, dst=dst, gateway=router)

    # remove the interface
    iface = context.ndb.interfaces[if_host].remove().commit()

    # check there is no interface, no route
    assert not interface_exists(context.netns, ifname=if_host)
    assert not interface_exists(context.netns, ifname=if_vlan)
    assert not address_exists(context.netns, ifname=if_vlan, address=ifaddr1)
    assert not address_exists(context.netns, ifname=if_vlan, address=ifaddr2)
    assert not route_exists(context.netns, dst=dst, gateway=router)

    # revert the changes using the implicit last_save
    iface.rollback()
    assert interface_exists(context.netns, ifname=if_host)
    assert interface_exists(context.netns, ifname=if_vlan)
    assert address_exists(context.netns, ifname=if_vlan, address=ifaddr1)
    assert address_exists(context.netns, ifname=if_vlan, address=ifaddr2)
    assert route_exists(context.netns, dst=dst, gateway=router)



import errno
from uuid import uuid4

import pytest
from pr2test.context_manager import NDBContextManager, SpecContextManager
from utils import require_user

from pyroute2.ipset import IPSet, IPSetError
from pyroute2.wiset import COUNT

pytest_plugins = "pytester"


@pytest.fixture
def context(request, tmpdir):
    '''
    This fixture is used to prepare the environment and
    to clean it up after each test.

    https://docs.pytest.org/en/stable/fixture.html
    '''
    #                                       test stage:
    #
    ctx = NDBContextManager(request, tmpdir)  # setup
    yield ctx  # execute
    ctx.teardown()  # cleanup


@pytest.fixture
def spec(request, tmpdir):
    '''
    A simple fixture with only some variables set
    '''
    ctx = SpecContextManager(request, tmpdir)
    yield ctx
    ctx.teardown()


@pytest.fixture
def ipset():
    require_user('root')
    sock = IPSet()
    yield sock
    sock.close()


@pytest.fixture
def ipset_name(ipset):
    name = str(uuid4())[:16]
    yield name
    try:
        ipset.destroy(name)
    except IPSetError as e:
        if e.code != errno.ENOENT:
            raise


@pytest.fixture(params=(None, IPSet))
def wiset_sock(request):
    if request.param is None:
        yield None
    else:
        before_count = COUNT["count"]
        with IPSet() as sock:
            yield sock
        assert before_count == COUNT['count']



import pyroute2


def test_exceptions():
    assert issubclass(pyroute2.NetlinkError, Exception)
    assert issubclass(pyroute2.CreateException, Exception)
    assert issubclass(pyroute2.CommitException, Exception)



from socket import AF_INET

import pytest
from pr2test.context_manager import make_test_matrix
from pr2test.marks import require_root

from pyroute2 import protocols
from pyroute2.netlink.rtnl import TC_H_INGRESS, TC_H_ROOT

pytestmark = [require_root()]
test_matrix = make_test_matrix(targets=['local', 'netns'])


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_simple(context):
    index, ifname = context.default_interface
    context.ipr.tc('add', 'ingress', index=index)
    qdisc = None
    for qdisc in context.ipr.get_qdiscs(index=index):
        if qdisc.get_attr('TCA_KIND') == 'ingress':
            break
    else:
        raise FileNotFoundError('qdisc not found')
    assert qdisc['handle'] == 0xFFFF0000
    assert qdisc['parent'] == TC_H_INGRESS
    return (index, ifname)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_filter(context):
    index, ifname = test_simple(context)
    context.ipr.tc(
        'add-filter',
        'u32',
        index=index,
        protocol=AF_INET,
        parent=0xFFFF0000,
        action='drop',
        target=0x1,
        rate='10kbit',
        burst=10240,
        limit=0,
        prio=50,
        keys=['0x0/0x0+12'],
    )
    fls = context.ipr.get_filters(index=index, parent=0xFFFF0000)
    # assert there are filters
    assert fls
    # assert there is one police rule:
    prs = [
        x
        for x in fls
        if x.get_attr('TCA_OPTIONS') is not None
        and (
            x.get_attr('TCA_OPTIONS').get_attr('TCA_U32_POLICE') is not None
            or x.get_attr('TCA_OPTIONS').get_attr('TCA_U32_ACT') is not None
        )
    ][0]
    # assert the police rule has specified parameters
    options = prs.get_attr('TCA_OPTIONS')
    police_u32 = options.get_attr('TCA_U32_POLICE')
    # on modern kernels there is no TCA_U32_POLICE under
    # TCA_OPTIONS, but there is TCA_U32_ACT
    if police_u32 is None:
        police_u32 = (
            options.get_attr('TCA_U32_ACT')
            .get_attr('TCA_ACT_PRIO_0')
            .get_attr('TCA_ACT_OPTIONS')
        )
    police_tbf = police_u32.get_attr('TCA_POLICE_TBF')
    assert police_tbf['rate'] == 1250
    assert police_tbf['mtu'] == 2040


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_action_stats(context):
    index, ifname = test_simple(context)
    context.ipr.tc(
        'add-filter',
        'u32',
        index=index,
        parent='ffff:',
        protocol=protocols.ETH_P_ALL,
        keys=['0x0/0x0+0'],
        target=TC_H_ROOT,
        action={'kind': 'gact', 'action:': 'ok'},
    )
    fls = context.ipr.get_filters(index=index, parent=TC_H_INGRESS)

    act = [
        x
        for x in fls
        if x.get_attr('TCA_OPTIONS') is not None
        and (x.get_attr('TCA_OPTIONS').get_attr('TCA_U32_ACT') is not None)
    ][0]
    # assert we have a u32 filter with a gact action
    assert act.get_attr('TCA_KIND') == 'u32'
    gact = (
        act.get_attr("TCA_OPTIONS")
        .get_attr("TCA_U32_ACT")
        .get_attr("TCA_ACT_PRIO_1")
    )
    assert gact.get_attr('TCA_ACT_KIND') == 'gact'
    # assert our gact has stats
    assert gact.get_attr('TCA_ACT_STATS')



from setup import lab
from task import IPRoute

if len(lab.registry) != 1:
    raise AssertionError('expected exactly one IPRoute instance')
if not isinstance(lab.registry[0], IPRoute):
    raise AssertionError('expected IPRoute instance')

ipr = lab.registry[0]
ipr.get_addr.assert_called()
ipr.close.assert_called_once()



from pr2test.marks import require_root

from pyroute2 import IPBatch

pytestmark = [require_root()]


def test_link_add(context):

    ifname = context.new_ifname

    ipb = IPBatch()
    ipb.link('add', ifname=ifname, kind='dummy')
    data = ipb.batch
    ipb.reset()
    ipb.close()
    context.ipr.sendto(data, (0, 0))
    context.ndb.interfaces.wait(ifname=ifname, timeout=3)



from pyroute2.iproute.linux import IPRoute
from pyroute2.nslink.nslink import NetNS


def interface_exists(netns=None, *argv, **kwarg):
    ret = 0
    ipr = None

    if netns is None:
        ipr = IPRoute()
    else:
        ipr = NetNS(netns)

    spec = {}
    spec.update(kwarg)
    ret = list(ipr.link_lookup(*argv, **spec))
    ipr.close()

    return len(ret) >= 1


def address_exists(netns=None, **kwarg):
    ret = 0
    ipr = None

    if netns is None:
        ipr = IPRoute()
    else:
        ipr = NetNS(netns)

    if 'match' in kwarg:
        nkw = kwarg['match']
    else:
        nkw = dict(kwarg)
        for key in ('address', 'local'):
            if key in nkw:
                nkw[key] = nkw[key].split('/')[0]

    if 'ifname' in kwarg:
        links = list(ipr.link_lookup(ifname=kwarg['ifname']))
        if links:
            nkw['index'] = links[0]
            nkw.pop('ifname')
        else:
            ipr.close()
            return 0

    ret = list(ipr.addr('dump', match=nkw))
    ipr.close()

    return len(ret) == 1


def rtnl_object_exists(api, netns, record_filter):
    ret = 0
    ipr = None

    if netns is None:
        ipr = IPRoute()
    else:
        ipr = NetNS(netns)

    ret = list(getattr(ipr, api)('dump', **record_filter))
    ipr.close()
    return len(ret) >= 1


def neighbour_exists(netns=None, **kwarg):
    return rtnl_object_exists('neigh', netns, kwarg)


def route_exists(netns=None, **kwarg):
    return rtnl_object_exists('route', netns, kwarg)


def rule_exists(netns=None, **kwarg):
    return rtnl_object_exists('rule', netns, kwarg)


def fdb_record_exists(netns=None, **kwarg):
    return rtnl_object_exists('fdb', netns, kwarg)


def qdisc_exists(netns=None, kind=None, **kwarg):
    if netns is None:
        ipr = IPRoute()
    else:
        ipr = NetNS(netns)
    opts = {}
    with ipr:
        if 'ifname' in kwarg:
            opts['index'] = ipr.link_lookup(ifname=kwarg.pop('ifname'))[0]
        ret = list(ipr.get_qdiscs(**opts))
        if kind is not None:
            ret = [x for x in ret if x.get_attr('TCA_KIND') == kind]
        if kwarg:
            pre = ret
            ret = []
            for qdisc in pre:
                options = qdisc.get_attr('TCA_OPTIONS')
                if 'attrs' in options:
                    options = dict(options['attrs'])
                for opt in kwarg:
                    if kwarg[opt] not in (
                        options.get(opt),
                        options.get(qdisc.name2nla(opt)),
                    ):
                        break
                else:
                    ret.append(qdisc)
        return ret



#!/usr/bin/env python

import pathlib
import sys

from jinja2 import Environment, FileSystemLoader

env = Environment(loader=FileSystemLoader('lab/_templates/'))
# js template
template = env.get_template('conf.js')
with open('lab/_static/conf.js', 'w') as f:
    f.write(template.render(distfile=sys.argv[1]))
    print('created lab/_static/conf.js')

# html template
template = env.get_template('form_template.html')
root = pathlib.Path('examples/lab')
for example in root.iterdir():
    setup = example.joinpath('setup.py').read_text()
    task = example.joinpath('task.py').read_text()
    check = ''
    with example.joinpath('check.py').open('r') as f:
        for line in f.readlines():
            if 'import' not in line:
                check += line
    name = example.name
    with open(f'lab/{name}.html', 'w') as f:
        f.write(
            template.render(setup=setup, task=task, check=check, name=name)
        )
        print(f'created lab/{name}.html')



from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg, protinfo_bridge

from .common import Index, IPRouteFilter, NLAKeyTransform


class BridgeFieldFilter(Index, NLAKeyTransform):

    _nla_prefix = ifinfmsg.prefix


class BridgeIPRouteFilter(IPRouteFilter):
    def finalize(self, context):
        if self.command != 'dump':
            for key in ('vlan_info', 'mode', 'vlan_flags'):
                if key in context:
                    if 'IFLA_AF_SPEC' not in context:
                        context['IFLA_AF_SPEC'] = {'attrs': []}
                    nla = ifinfmsg.af_spec_bridge.name2nla(key)
                    value = context[key]
                    try:
                        del context[key]
                    except KeyError:
                        pass
                    context['IFLA_AF_SPEC']['attrs'].append([nla, value])


class BridgePortFieldFilter(IPRouteFilter):

    _nla_prefix = ifinfmsg.prefix
    _allowed = [x[0] for x in protinfo_bridge.nla_map]
    _allowed.append('attrs')

    def finalize(self, context):
        keys = tuple(context.keys())
        context['attrs'] = []
        for key in keys:
            context['attrs'].append(
                (protinfo_bridge.name2nla(key), context[key])
            )



import pytest
from pr2test.context_manager import make_test_matrix, skip_if_not_supported
from pr2test.marks import require_root

from pyroute2 import NetlinkError
from pyroute2.netlink.rtnl.ifinfmsg import IFF_NOARP

pytestmark = [require_root()]

test_matrix = make_test_matrix(targets=['local', 'netns'])


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_updown_link(context):
    index, ifname = context.default_interface

    context.ipr.link('set', index=index, state='up')
    assert context.ipr.get_links(ifname=ifname)[0]['flags'] & 1
    context.ipr.link('set', index=index, state='down')
    assert not (context.ipr.get_links(ifname=ifname)[0]['flags'] & 1)


@skip_if_not_supported
@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_link_altname(context):
    altname1 = context.new_ifname
    altname2 = context.new_ifname
    weird_name = "test_with_a_very_long_string" "_and_♄⚕⚚_utf8_symbol"
    index, ifname = context.default_interface

    for name in (altname1, altname2, weird_name):
        with pytest.raises(NetlinkError):
            context.ipr.link("get", altname=name)

    context.ipr.link("property_add", index=index, altname=[altname1, altname2])
    assert len(context.ipr.link("get", altname=altname1)) == 1
    assert len(context.ipr.link("get", altname=altname2)) == 1

    context.ipr.link("property_del", index=index, altname=[altname1, altname2])

    for name in (altname1, altname2):
        with pytest.raises(NetlinkError):
            context.ipr.link("get", altname=name)

    context.ipr.link("property_add", index=index, altname=weird_name)
    assert len(context.ipr.link("get", altname=weird_name)) == 1
    context.ipr.link("property_del", index=index, altname=weird_name)
    assert len(context.ipr.link("dump", altname=weird_name)) == 0
    with pytest.raises(NetlinkError):
        context.ipr.link("get", altname=weird_name)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_link_filter(context):
    links = context.ipr.link('dump', ifname='lo')
    assert len(links) == 1
    assert links[0].get_attr('IFLA_IFNAME') == 'lo'


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_link_legacy_nla(context):
    index, ifname = context.default_interface
    new_ifname = context.new_ifname

    context.ipr.link('set', index=index, state='down')
    context.ipr.link('set', index=index, IFLA_IFNAME=new_ifname)
    assert context.ipr.link_lookup(ifname=new_ifname) == [index]

    context.ipr.link('set', index=index, ifname=ifname)
    assert context.ipr.link_lookup(ifname=ifname) == [index]


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_link_rename(context):
    index, ifname = context.default_interface
    new_ifname = context.new_ifname
    context.ndb.interfaces[ifname].set('state', 'down').commit()

    context.ipr.link('set', index=index, ifname=new_ifname)
    assert context.ipr.link_lookup(ifname=new_ifname) == [index]

    context.ipr.link('set', index=index, ifname=ifname)
    assert context.ipr.link_lookup(ifname=ifname) == [index]


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_link_arp_flag(context):
    index, _ = context.default_interface

    # by default dummy interface have NOARP set
    assert context.ipr.get_links(index)[0]['flags'] & IFF_NOARP

    context.ipr.link('set', index=index, arp=True)
    assert not context.ipr.get_links(index)[0]['flags'] & IFF_NOARP

    context.ipr.link('set', index=index, arp=False)
    assert context.ipr.get_links(index)[0]['flags'] & IFF_NOARP

    context.ipr.link('set', index=index, noarp=False)
    assert not context.ipr.get_links(index)[0]['flags'] & IFF_NOARP

    context.ipr.link('set', index=index, noarp=True)
    assert context.ipr.get_links(index)[0]['flags'] & IFF_NOARP


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_symbolic_flags_ifinfmsg(context):
    index, _ = context.default_interface

    context.ipr.link('set', index=index, flags=['IFF_UP'])
    iface = context.ipr.get_links(index)[0]
    assert iface['flags'] & 1
    assert 'IFF_UP' in iface.flags2names(iface['flags'])
    context.ipr.link('set', index=index, flags=['!IFF_UP'])
    assert not (context.ipr.get_links(index)[0]['flags'] & 1)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_remove_link(context):
    index, ifname = context.default_interface
    context.ipr.link('del', index=index)
    assert len(context.ipr.link_lookup(ifname=ifname)) == 0
    assert len(context.ipr.link_lookup(index=index)) == 0


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_brport_basic(context):

    bridge = context.new_ifname
    port = context.new_ifname

    context.ndb.interfaces.create(
        ifname=bridge, kind='bridge', state='up'
    ).commit()
    context.ndb.interfaces.create(
        ifname=port, kind='dummy', state='up'
    ).commit()

    context.ipr.link(
        'set',
        index=context.ndb.interfaces[port]['index'],
        master=context.ndb.interfaces[bridge]['index'],
    )

    context.ipr.brport(
        'set',
        index=context.ndb.interfaces[port]['index'],
        unicast_flood=0,
        cost=200,
        proxyarp=1,
    )

    port = context.ipr.brport(
        'dump', index=context.ndb.interfaces[port]['index']
    )[0]
    protinfo = port.get_attr('IFLA_PROTINFO')
    assert protinfo.get_attr('IFLA_BRPORT_COST') == 200
    assert protinfo.get_attr('IFLA_BRPORT_PROXYARP') == 1
    assert protinfo.get_attr('IFLA_BRPORT_UNICAST_FLOOD') == 0



import time
from socket import AF_INET, AF_INET6

import pytest
from pr2test.context_manager import skip_if_not_supported
from pr2test.marks import require_root

pytestmark = [require_root()]


def test_real_links(context):
    links = set([x['index'] for x in context.ipr.get_links()])
    neigh = set([x['ifindex'] for x in context.ipr.get_neighbours()])
    assert neigh <= links


def test_filter(context):
    ipaddr1 = context.new_ipaddr
    ipaddr2 = context.new_ipaddr
    index, ifname = context.default_interface
    lladdr = '00:11:22:33:44:55'
    # this is required -- the default interface takes time to setup
    time.sleep(0.5)
    # inject arp records
    context.ipr.neigh('add', dst=ipaddr1, lladdr=lladdr, ifindex=index)
    context.ipr.neigh('add', dst=ipaddr2, lladdr=lladdr, ifindex=index)
    # assert two arp records on the interface
    assert len(context.ipr.get_neighbours(ifindex=index, family=AF_INET)) == 2
    # filter by dst
    assert len(context.ipr.get_neighbours(dst=ipaddr1)) == 1
    # filter with lambda
    assert (
        len(
            context.ipr.get_neighbours(
                match=lambda x: x['ifindex'] == index
                and x['family'] == AF_INET
            )
        )
        == 2
    )


@skip_if_not_supported
def test_get(context):
    ipaddr1 = context.new_ipaddr
    index, ifname = context.default_interface
    lladdr = '00:11:22:33:44:55'
    # this is required -- the default interface takes time to setup
    time.sleep(0.5)
    context.ipr.neigh('add', dst=ipaddr1, lladdr=lladdr, ifindex=index)
    res = context.ipr.neigh('get', dst=ipaddr1, ifindex=index)
    assert res[0].get_attr("NDA_DST") == ipaddr1


@pytest.mark.parametrize(
    'family,ipaddr_source,prefixlen',
    ((AF_INET, 'new_ipaddr', 24), (AF_INET6, 'new_ip6addr', 64)),
)
def test_dump(context, family, ipaddr_source, prefixlen):
    index, ifname = context.default_interface
    ipaddr1 = getattr(context, ipaddr_source)
    # wait for the link
    context.ipr.poll(context.ipr.link, 'dump', index=index, state='up')
    context.ipr.addr(
        'add', index=index, family=family, address=ipaddr1, prefixlen=prefixlen
    )
    # wait for the address
    context.ipr.poll(context.ipr.addr, 'dump', index=index, address=ipaddr1)
    # now add neighbours; to keep it simpler, don't take care if we loose
    # some of the neighbour records, enough to have there at least one, so
    # add some and continue
    for last_byte in range(32):
        l2addr = f'00:11:22:33:44:{last_byte:02}'
        context.ipr.neigh(
            'add',
            dst=getattr(context, ipaddr_source),
            lladdr=l2addr,
            ifindex=index,
        )
    # ok, now the dump should not be empty
    assert len(list(context.ipr.neigh('dump', family=family))) > 0



import contextlib
import io
import sys

ret = io.StringIO()

with contextlib.ExitStack() as ctx:
    map_file = ctx.enter_context(open(sys.argv[1], 'r'))
    img_file = ctx.enter_context(open(sys.argv[2], 'r'))

    mapping = {
        key.strip(): value.strip()
        for (key, value) in [x.split('|') for x in map_file.readlines()]
    }
    for line in img_file.readlines():
        for key, value in mapping.items():
            line = line.replace(key, f'  <a href="{value}">{key}</a>')
        ret.write(line)

with open(sys.argv[2], 'w') as img_file:
    img_file.write(ret.getvalue())



from socket import AF_INET6

import pytest
from pr2test.context_manager import make_test_matrix
from pr2test.marks import require_root
from pr2test.tools import rule_exists

pytestmark = [require_root()]

test_matrix = make_test_matrix(
    targets=['local', 'netns'],
    tables=[100, 10000],
    dbs=['sqlite3/:memory:', 'postgres/pr2test'],
)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_explicit_ipv6_src(context):
    ipnet = context.new_ip6net
    table = context.table

    spec = {
        'family': AF_INET6,
        'src': ipnet.network,
        'src_len': ipnet.netmask,
        'table': table,
        'priority': 50,
    }
    context.register_rule(spec)

    context.ndb.rules.create(**spec).commit()
    assert rule_exists(context.netns, **spec)

    context.ndb.rules[spec].remove().commit()
    assert not rule_exists(context.netns, **spec)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_implicit_ipv6_src(context):
    ipnet = context.new_ip6net
    table = context.table

    spec = {
        'src': ipnet.network,
        'src_len': ipnet.netmask,
        'table': table,
        'priority': 50,
    }
    search_spec = spec.copy()
    search_spec['family'] = AF_INET6
    context.register_rule(search_spec)

    context.ndb.rules.create(**spec).commit()
    assert rule_exists(context.netns, **search_spec)

    context.ndb.rules[spec].remove().commit()
    assert not rule_exists(context.netns, **search_spec)



import resource
import socket

import pytest

from pyroute2.netlink.nlsocket import NetlinkSocket


def test_ports_auto():
    # create two sockets
    s1 = NetlinkSocket()
    s2 = NetlinkSocket()

    # both bind() should succeed
    s1.bind()
    s2.bind()

    # check that ports are different
    assert s1.port != s2.port

    s1.close()
    s2.close()


def test_ports_fail():
    s1 = NetlinkSocket(port=0x10)
    s2 = NetlinkSocket(port=0x10)

    # check if ports are set
    assert s1.port == s2.port

    # bind the first socket, must succeed
    s1.bind()

    # bind the second, must fail
    exception = None
    with pytest.raises(socket.error) as exception:
        s2.bind()
    # socket.error / OSError(98, 'Address already in use')
    assert exception.value.errno == 98

    s1.close()
    s2.close()


def test_no_free_ports():
    soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
    try:
        resource.setrlimit(resource.RLIMIT_NOFILE, (4096, 4096))
    except ValueError:
        pytest.skip('cannot set RLIMIT_NOFILE')

    # create and bind 1024 sockets
    ports = [NetlinkSocket() for x in range(1024)]
    for port in ports:
        port.bind()

    # create an extra socket
    fail = NetlinkSocket()
    # bind must fail with KeyError: no free ports available
    with pytest.raises(KeyError):
        fail.bind()

    # cleanup
    for port in ports:
        port.close()

    fail.close()
    resource.setrlimit(resource.RLIMIT_NOFILE, (soft, hard))



from time import sleep

from pr2test.marks import require_root
from utils import require_kernel

from pyroute2.netlink.exceptions import IPSetError
from pyroute2.wiset import (
    IPStats,
    WiSet,
    get_ipset_socket,
    load_all_ipsets,
    load_ipset,
)
from pyroute2.wiset import test_ipset_exist as ipset_exists

pytestmark = [require_root()]


def test_create_one_ipset(ipset_name, wiset_sock):
    with WiSet(name=ipset_name, sock=wiset_sock) as myset:
        myset.create()

        list_wiset = load_all_ipsets(sock=wiset_sock)
        assert ipset_exists(ipset_name, sock=wiset_sock)
        myset.destroy()

        assert not ipset_exists(ipset_name, sock=wiset_sock)
        assert ipset_name in list_wiset
        assert ipset_name not in load_all_ipsets(sock=wiset_sock)


def test_create_ipset_twice(ipset_name, wiset_sock):
    with WiSet(name=ipset_name, sock=wiset_sock) as myset:
        myset.create()

        try:
            myset.create(exclusive=True)
            assert False
        except IPSetError:
            pass

        myset.create(exclusive=False)
        myset.destroy()
        assert ipset_name not in load_all_ipsets(sock=wiset_sock)


def test_check_ipset_stats(ipset_name, wiset_sock):
    def test_stats(myset, res=None, counters=False):
        myset.counters = counters
        myset.create()
        myset.add("8.8.8.8", packets=res, bytes=res)
        myset.update_content()
        stats = myset.content
        myset.flush()

        assert stats["8.8.8.8"].packets == res
        assert stats["8.8.8.8"].bytes == res

        myset.add("8.8.8.8")
        myset.update_content()
        stats = myset.content
        if res is not None:
            res = 0
        assert stats["8.8.8.8"].packets == res
        assert stats["8.8.8.8"].bytes == res
        myset.destroy()

    with WiSet(name=ipset_name, sock=wiset_sock) as myset:
        test_stats(myset, res=10, counters=True)
        test_stats(myset)


def test_ipset_with_comment(ipset_name, wiset_sock):
    comment = "test comment"

    with WiSet(name=ipset_name, sock=wiset_sock, comment=True) as myset:
        myset.create()
        myset.add("8.8.8.8", comment=comment)
        set_list = myset.content

    assert set_list["8.8.8.8"].comment == comment


def test_ipset_with_skbinfo(ipset_name, wiset_sock):
    with WiSet(name=ipset_name, sock=wiset_sock, skbinfo=True) as myset:
        myset.create()
        myset.add("192.168.1.1", skbmark=(0xC8, 0xC8))
        myset.add("192.168.1.2", skbmark=(0xC9, 0xFFFFFFFF))
        myset.add("192.168.1.3", skbmark="0xca/0xca")
        myset.add("192.168.1.4", skbmark="0xCB")
        set_list = myset.content
        myset.destroy()

    assert set_list["192.168.1.1"].skbmark == "0xc8/0xc8"
    assert set_list["192.168.1.2"].skbmark == "0xc9"
    assert set_list["192.168.1.3"].skbmark == "0xca/0xca"
    assert set_list["192.168.1.4"].skbmark == "0xcb"


def test_list_on_large_set(ipset_name, wiset_sock):
    set_size = 30000
    base_ip = "10.10.%d.%d"

    with WiSet(name=ipset_name, sock=wiset_sock) as myset:
        myset.create()
        for i in range(0, set_size):
            myset.add(base_ip % (i / 255, i % 255))
        stats_len = len(myset.content)
        stats_len2 = len(
            load_ipset(ipset_name, content=True, sock=wiset_sock).content
        )

    assert stats_len == set_size
    assert stats_len2 == set_size


def test_remove_entry(ipset_name, wiset_sock):
    ip = "1.1.1.1"

    with WiSet(name=ipset_name, sock=wiset_sock, counters=True) as myset:
        myset.create()
        myset.add(ip)
        assert ip in myset.content
        myset.delete(ip)
        myset.update_content()
        assert ip not in myset.content


def test_flush(ipset_name, wiset_sock):
    ip_list = ["1.2.3.4", "1.1.1.1", "7.7.7.7"]

    with WiSet(name=ipset_name, sock=wiset_sock) as myset:
        myset.create()
        for ip in ip_list:
            myset.add(ip)
            assert myset.test(ip)
        myset.flush()
        for ip in ip_list:
            assert not myset.test(ip)


def test_list_in(ipset_name, wiset_sock):
    ip_list_good = ["1.2.3.4", "1.1.1.1", "7.7.7.7"]
    ip_list_bad = ["4.4.4.4", "5.5.5.5", "6.6.6.6"]

    with WiSet(name=ipset_name, sock=wiset_sock) as myset:
        myset.create()
        myset.replace_entries(ip_list_good)
        res_test = myset.test_list(ip_list_good + ip_list_bad)
        for ip in ip_list_good:
            assert ip in res_test
        for ip in ip_list_bad:
            assert ip not in res_test


def test_timeout(ipset_name, wiset_sock):
    ip = "1.2.3.4"
    timeout = 2

    with WiSet(name=ipset_name, sock=wiset_sock, timeout=timeout) as myset:
        myset.create()
        myset.add(ip)
        sleep(3)
        myset.update_content()
        assert ip not in myset.content
        assert timeout == load_ipset(ipset_name, sock=wiset_sock).timeout
        myset.add(ip, timeout=0)
        myset.update_content()
        assert myset.content[ip].timeout == 0


def test_basic_attribute_reads(ipset_name, wiset_sock):
    for value in [True, False]:
        myset = WiSet(
            name=ipset_name, sock=wiset_sock, counters=value, comment=value
        )
        if wiset_sock is None:
            myset.open_netlink()
        myset.create()
        from_netlink = load_ipset(ipset_name, sock=wiset_sock)
        assert value == from_netlink.comment
        assert value == from_netlink.counters
        myset.destroy()
        if wiset_sock is None:
            myset.close_netlink()


def test_replace_content(ipset_name, wiset_sock):
    list_a = ["1.2.3.4", "2.3.4.5", "6.7.8.9"]
    list_b = ["1.1.1.1", "2.2.2.2", "3.3.3.3"]

    def test_replace(content_a, content_b):
        myset = WiSet(name=ipset_name, sock=wiset_sock)
        myset.create()
        myset.insert_list(content_a)
        myset.update_content()
        for value in content_a:
            assert value in myset.content
        myset.replace_entries(content_b)
        myset.update_content()
        for (old, new) in zip(content_a, content_b):
            assert old not in myset.content
            assert new in myset.content
        myset.destroy()

    test_replace(list_a, list_b)
    test_replace(set(list_a), set(list_b))


def test_replace_content_with_comment(ipset_name, wiset_sock):
    list_a = [
        {'entry': "1.2.3.4", 'comment': 'foo'},
        {'entry': "2.3.4.5", 'comment': 'foo'},
        {'entry': "6.7.8.9", 'comment': 'bar'},
    ]
    list_b = [
        {'entry': "1.1.1.1", 'comment': 'foo'},
        {'entry': "2.2.2.2", 'comment': 'bar'},
        {'entry': "3.3.3.3", 'comment': 'foo'},
    ]

    def test_replace(content_a, content_b):
        myset = WiSet(name=ipset_name, sock=wiset_sock, comment=True)
        myset.create()
        myset.insert_list(content_a)
        myset.update_content()
        for value in content_a:
            assert value['entry'] in myset.content
            assert value['comment'] == (myset.content[value['entry']].comment)
        myset.replace_entries(content_b)
        myset.update_content()
        for value in content_a:
            assert value['entry'] not in myset.content
        for value in content_b:
            assert value['entry'] in myset.content
            assert value['comment'] == (myset.content[value['entry']].comment)

    test_replace(list_a, list_b)


def test_hash_net_ipset(ipset_name, wiset_sock):
    to_add = ["192.168.1.0/24", "192.168.2.0/23", "10.0.0.0/8"]
    atype = "hash:net"

    with WiSet(name=ipset_name, attr_type=atype, sock=wiset_sock) as myset:
        myset.create()
        myset.insert_list(to_add)
        for value in to_add:
            assert value in myset.content


def test_two_dimensions_ipset(ipset_name, wiset_sock):
    to_add = ["192.168.1.0/24,eth0", "192.168.2.0/23,eth1", "10.0.0.0/8,tun0"]
    atype = "hash:net,iface"

    with WiSet(name=ipset_name, attr_type=atype, sock=wiset_sock) as myset:
        myset.create()
        myset.insert_list(to_add)
        for value in to_add:
            assert value in myset.content


def test_stats_consistency(ipset_name, wiset_sock):
    """Test several way to fill the statistics of one IPSet"""
    entries = ["1.2.3.4", "1.2.3.5", "1.2.3.6"]

    myset = WiSet(name=ipset_name, sock=wiset_sock)
    myset.create()
    myset.insert_list(entries)

    myset_lists = load_all_ipsets(sock=wiset_sock, content=True)[ipset_name]
    for value in entries:
        assert value in myset_lists.content

    myset_list = load_ipset(ipset_name, sock=wiset_sock, content=True)
    for value in entries:
        assert value in myset_list.content


def test_hashnet_with_comment(ipset_name, wiset_sock):
    comment = "abcdef"
    myset = WiSet(
        name=ipset_name, attr_type="hash:net", comment=True, sock=wiset_sock
    )
    myset.create()

    inherit_sock = wiset_sock is not None
    myset = load_ipset(ipset_name, sock=wiset_sock, inherit_sock=inherit_sock)
    assert myset.comment

    myset.add("192.168.1.1", comment=comment)
    myset.update_content()

    assert myset.content["192.168.1.1/32"].comment == comment


def test_add_ipstats(ipset_name, wiset_sock):
    data = IPStats(
        packets=10,
        bytes=1000,
        comment="hello world",
        skbmark="0x10/0x10",
        timeout=None,
    )
    myset = WiSet(
        name=ipset_name,
        attr_type="hash:net",
        comment=True,
        skbinfo=True,
        counters=True,
        sock=wiset_sock,
    )
    myset.create()
    myset.add("198.51.100.0/24", **data._asdict())

    assert "198.51.100.0/24" in myset.content
    assert data == myset.content["198.51.100.0/24"]


def test_revision(ipset_name, wiset_sock):
    myset = WiSet(name=ipset_name, attr_type="hash:net", sock=wiset_sock)

    myset.create()
    assert load_ipset(ipset_name, sock=wiset_sock).revision >= 6


def test_force_attr_revision(ipset_name, wiset_sock):
    sock = get_ipset_socket(attr_revision=2)

    myset = WiSet(name=ipset_name, attr_type="hash:net", sock=wiset_sock)
    myset.create()
    assert load_ipset(ipset_name, sock=wiset_sock).revision >= 2
    sock.close()


def test_physdev(ipset_name):
    myset = WiSet(name=ipset_name, attr_type="hash:net,iface")
    myset.create()
    myset.add("192.168.0.0/24,eth0", physdev=False)
    myset.add("192.168.1.0/24,eth0", physdev=True)

    content = myset.content

    assert content["192.168.0.0/24,eth0"].physdev is False
    assert content["192.168.1.0/24,eth0"].physdev is True


def test_wildcard_entries(ipset_name):
    require_kernel(5, 5)
    myset = WiSet(name=ipset_name, attr_type="hash:net,iface")
    myset.create()
    myset.add("192.168.0.0/24,eth", wildcard=True)
    myset.add("192.168.1.0/24,wlan0", wildcard=False)

    content = myset.content

    assert content["192.168.0.0/24,eth"].wildcard is True
    assert content["192.168.1.0/24,wlan0"].wildcard is False


def test_invalid_load_ipset():
    assert load_ipset("ipsetdoesnotexists") is None



from pr2test.context_manager import skip_if_not_supported

from pyroute2 import IPRoute, RemoteIPRoute


@skip_if_not_supported
def test_links():
    with IPRoute() as ipr:
        links1 = set([x.get_attr('IFLA_IFNAME') for x in ipr.get_links()])

    with RemoteIPRoute() as ipr:
        links2 = set([x.get_attr('IFLA_IFNAME') for x in ipr.get_links()])

    assert links1 == links2



import pytest

from pyroute2 import IPRoute


class BasicContextManager:
    def __init__(self, request, tmpdir):
        self.ipr = IPRoute()

    def teardown(self):
        self.ipr.close()


@pytest.fixture
def context(request, tmpdir):
    ctx = BasicContextManager(request, tmpdir)
    yield ctx
    ctx.teardown()



import getpass
import json
import os
import subprocess
import sys

import nox

nox.options.envdir = f'./.nox-{getpass.getuser()}'
nox.options.reuse_existing_virtualenvs = False
nox.options.sessions = [
    'linter',
    'repo',
    'unit',
    'lab',
    'neutron',
    'integration',
    'linux-3.6',
    'linux-3.10',
    'minimal',
]

linux_kernel_modules = [
    'dummy',
    'bonding',
    '8021q',
    'mpls_router',
    'mpls_iptunnel',
    'l2tp_ip',
    'l2tp_eth',
    'l2tp_netlink',
]


def add_session_config(func):
    '''Decorator to load the session config.

    Usage::

        @nox.session
        @load_session_config
        def my_session_func(session, config):
            pass

    Command line usage::

        nox -e my_session_name -- '{"option": value}'

    The session config must be a valid JSON dictionary of options.
    '''

    def wrapper(session):
        if session.posargs:
            config = json.loads(session.posargs[0])
        else:
            config = {}
        session.debug(f'session config: {config}')
        return func(session, config)

    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    wrapper.__has_user_config__ = True
    return wrapper


def options(module, config):
    '''Return pytest options set.'''
    ret = [
        'python',
        '-m',
        'pytest',
        '--basetemp',
        './log',
        '--exitfirst',
        '--verbose',
        '--junitxml=junit.xml',
    ]
    if config.get('pdb'):
        ret.append('--pdb')
    if config.get('coverage'):
        ret.append('--cov-report=html')
        ret.append('--cov=pyroute2')
    if config.get('tests_prefix'):
        module = f'{config["tests_prefix"]}/{module}'
    if config.get('sub'):
        module = f'{module}/{config["sub"]}'
    ret.append(module)
    return ret


def setup_linux(session):
    '''Setup a Linux system.

    Load all the modules, but ignore any errors: missing kernel API
    will be handled at the test module level. Same for sysctl.
    '''
    if sys.platform == 'linux' and getpass.getuser() == 'root':
        for module in linux_kernel_modules:
            session.run(
                'modprobe', module, external=True, success_codes=[0, 1]
            )
        session.run(
            'sysctl',
            'net.mpls.platform_labels=2048',
            external=True,
            success_codes=[0, 255],
        )


def setup_venv_minimal(session, config):
    if not config.get('reuse'):
        session.install('--upgrade', 'pip')
        session.install('build')
        session.install('twine')
        session.install('-r', 'requirements.dev.txt')
        session.install('-r', 'requirements.docs.txt')
        session.run('mv', '-f', 'setup.cfg', 'setup.cfg.orig', external=True)
        session.run('cp', 'setup.minimal.cfg', 'setup.cfg', external=True)
        session.run('python', '-m', 'build')
        session.run('python', '-m', 'twine', 'check', 'dist/*')
        session.install('.')
        session.run('mv', '-f', 'setup.cfg.orig', 'setup.cfg', external=True)
    tmpdir = os.path.abspath(session.create_tmp())
    session.run('cp', '-a', 'lab', tmpdir, external=True)
    session.run('cp', '-a', 'tests', tmpdir, external=True)
    session.run('cp', '-a', 'examples', tmpdir, external=True)
    return tmpdir


def setup_venv_common(session, flavour='dev'):
    session.install('--upgrade', 'pip')
    session.install('-r', f'requirements.{flavour}.txt')
    session.install('.')
    return os.path.abspath(session.create_tmp())


def setup_venv_dev(session):
    tmpdir = setup_venv_common(session)
    session.run('cp', '-a', 'tests', tmpdir, external=True)
    session.run('cp', '-a', 'examples', tmpdir, external=True)
    session.chdir(f'{tmpdir}/tests')
    return tmpdir


def setup_venv_repo(session):
    tmpdir = setup_venv_common(session, 'repo')
    for item in (
        ('tests', tmpdir),
        ('noxfile.py', tmpdir),
        ('VERSION', tmpdir),
        ('CHANGELOG.rst', tmpdir),
    ):
        session.run('cp', '-a', *item, external=True)
    git_ls_files = subprocess.run(
        ['git', 'ls-files', 'requirements*'], stdout=subprocess.PIPE
    )
    files = [x.decode('utf-8') for x in git_ls_files.stdout.split()]
    for fname in files:
        session.run('cp', '-a', fname, tmpdir, external=True)
    session.chdir(tmpdir)
    return tmpdir


def setup_venv_docs(session):
    tmpdir = setup_venv_common(session, 'docs')
    session.run('cp', '-a', 'docs', tmpdir, external=True)
    session.run('cp', '-a', 'examples', tmpdir, external=True)
    [
        session.run('cp', src, dst, external=True)
        for (src, dst) in (
            ('README.rst', f'{tmpdir}/docs/general.rst'),
            ('README.report.rst', f'{tmpdir}/docs/report.rst'),
            ('README.contribute.rst', f'{tmpdir}/docs/devcontribute.rst'),
            ('CHANGELOG.rst', f'{tmpdir}/docs/changelog.rst'),
        )
    ]
    return tmpdir


@nox.session(name='test-platform')
def test_platform(session):
    '''Test platform capabilities. Requires root to run.'''
    setup_venv_common(session)
    session.run('pyroute2-test-platform')


@nox.session
def docs(session):
    '''Generate project docs.'''
    tmpdir = setup_venv_docs(session)
    cwd = os.path.abspath(os.getcwd())
    # man pages
    session.chdir(f'{tmpdir}/docs/')
    session.run('make', 'man', external=True)
    session.run('cp', '-a', 'man', f'{cwd}/docs/', external=True)
    # html
    session.chdir(f'{tmpdir}/docs/')
    session.run('make', 'html', external=True)
    session.run('cp', '-a', 'html', f'{cwd}/docs/', external=True)
    session.chdir(cwd)
    session.run('bash', 'util/aafigure_mapper.sh', external=True)
    #
    session.log('8<---------------------------------------------------------')
    session.log('compiled docs:')
    session.log(f'html pages -> {cwd}/docs/html')
    session.log(f'man pages -> {cwd}/docs/man')


@nox.session
def linter(session):
    '''Run code checks and linters.'''
    session.install('pre-commit')
    session.run('pre-commit', 'run', '-a')


@nox.session
@add_session_config
def unit(session, config):
    '''Run unit tests.'''
    setup_venv_dev(session)
    session.run(*options('test_unit', config))


@nox.session
@add_session_config
def integration(session, config):
    '''Run integration tests (lnst, kuryr, ...).'''
    setup_venv_dev(session)
    session.run(*options('test_integration', config))


@nox.session(python=['3.6', '3.10'])
@add_session_config
def linux(session, config):
    '''Run Linux functional tests. Requires root to run all the tests.'''
    setup_linux(session)
    workspace = setup_venv_dev(session)
    session.run(
        *options('test_linux', config),
        env={
            'WORKSPACE': workspace,
            'SKIPDB': 'postgres',
            'PYTHONPATH': f'{workspace}/tests/mocklib',
        },
    )


@nox.session
@add_session_config
def minimal(session, config):
    '''Run tests on pyroute2.minimal package.'''
    tmpdir = setup_venv_minimal(session, config)
    session.chdir(f'{tmpdir}/tests')
    session.run(*options('test_minimal', config))


@nox.session
@add_session_config
def lab(session, config):
    '''Test lab code blocks.'''
    workspace = setup_venv_minimal(session, config)
    for fname in os.listdir('dist'):
        if fname.startswith('pyroute2.minimal') and fname.endswith('whl'):
            break
    session.run('python', 'util/make_lab_templates.py', fname, external=True)
    session.run('make', '-C', 'lab', 'html', external=True)
    session.run('cp', f'dist/{fname}', 'lab/_build/html/', external=True)
    # make tests
    session.chdir(f'{workspace}/tests')
    session.run(*options('test_lab', config), env={'WORKSPACE': workspace})


@nox.session
@add_session_config
def openbsd(session, config):
    '''Run OpenBSD tests. Requires OpenBSD >= 7.1'''
    setup_venv_dev(session)
    session.run(*options('test_openbsd', config))


@nox.session
@add_session_config
def neutron(session, config):
    '''Run Neutron integration tests.'''
    setup_venv_dev(session)
    session.run(*options('test_neutron', config))


@nox.session
@add_session_config
def repo(session, config):
    '''Run repo tests.'''
    setup_venv_repo(session)
    config['tests_prefix'] = 'tests'
    session.run(*options('test_repo', config))


@nox.session
def build(session):
    '''Run package build.'''
    session.install('build')
    session.install('twine')
    session.run('python', '-m', 'build')
    session.run('python', '-m', 'twine', 'check', 'dist/*')



import csv
import json
from socket import AF_INET

import pytest
from pr2test.context_manager import make_test_matrix, skip_if_not_supported
from pr2test.marks import require_root

from pyroute2.common import basestring
from pyroute2.ndb.objects import RTNL_Object
from pyroute2.ndb.report import Record, RecordSet

pytestmark = [require_root()]
test_matrix = make_test_matrix(
    targets=['local', 'netns'], dbs=['sqlite3/:memory:', 'postgres/pr2test']
)


@pytest.mark.parametrize(
    'view,key,item',
    (
        ('interfaces', 'ifname', 'lo'),
        ('routes', 'dst', '127.0.0.0/8'),
        ('addresses', 'address', '127.0.0.1/8'),
    ),
)
@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_contains(context, view, key, item):
    context.ndb.interfaces['lo'].set('state', 'up').commit()
    getattr(context.ndb, view).wait(**{key: item, 'timeout': 10})
    assert item in getattr(context.ndb, view)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_types(context):
    # check for the report type here
    assert isinstance(context.ndb.interfaces.summary(), RecordSet)
    # repr must be a string
    assert isinstance(repr(context.ndb.interfaces.summary()), basestring)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_iter_keys(context):
    for name in ('interfaces', 'addresses', 'neighbours', 'routes', 'rules'):
        view = getattr(context.ndb, name)
        for key in view:
            assert isinstance(key, Record)
            obj = view.get(key)
            if obj is not None:
                assert isinstance(obj, RTNL_Object)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_join(context):

    ifname = context.new_ifname
    ipaddr1 = context.new_ipaddr
    ipaddr2 = context.new_ipaddr

    (
        context.ndb.interfaces.create(ifname=ifname, kind='dummy', state='up')
        .add_ip(address=ipaddr1, prefixlen=24)
        .add_ip(address=ipaddr2, prefixlen=24)
        .commit()
    )

    addr = (
        context.ndb.addresses.dump()
        .filter(lambda x: x.family == AF_INET)
        .join(
            (context.ndb.interfaces.dump().filter(lambda x: x.state == 'up')),
            condition=lambda l, r: l.index == r.index and r.ifname == ifname,
            prefix='if_',
        )
        .select('address')
    )

    s1 = set((ipaddr1, ipaddr2))
    s2 = set([x.address for x in addr])
    assert s1 == s2


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_slices(context):
    a = list(context.ndb.rules.dump())
    ln = len(a) - 1
    # simple indices
    assert a[0] == context.ndb.rules.dump()[0]
    assert a[1] == context.ndb.rules.dump()[1]
    assert a[-1] == context.ndb.rules.dump()[-1]
    assert context.ndb.rules.dump()[ln] == a[-1]
    try:
        context.ndb.rules.dump()[len(a)]
    except IndexError:
        pass
    # slices
    assert a[0:] == context.ndb.rules.dump()[0:]
    assert a[:3] == context.ndb.rules.dump()[:3]
    assert a[0:3] == context.ndb.rules.dump()[0:3]
    assert a[1:3] == context.ndb.rules.dump()[1:3]
    # negative slices
    assert a[-3:] == context.ndb.rules.dump()[-3:]
    assert a[-3:-1] == context.ndb.rules.dump()[-3:-1]
    # mixed
    assert a[-ln : ln - 1] == context.ndb.rules.dump()[-ln : ln - 1]
    # step
    assert a[2:ln:2] == context.ndb.rules.dump()[2:ln:2]


@pytest.mark.parametrize('context', test_matrix, indirect=True)
@skip_if_not_supported
def test_report_chains(context):
    ipnet = str(context.ipnets[1].network)
    ipaddr = context.new_ipaddr
    router = context.new_ipaddr
    ifname = context.new_ifname

    (
        context.ndb.interfaces.create(ifname=ifname, kind='dummy', state='up')
        .add_ip(address=ipaddr, prefixlen=24)
        .commit()
    )

    (
        context.ndb.routes.create(
            dst=ipnet,
            dst_len=24,
            gateway=router,
            encap={'type': 'mpls', 'labels': [20, 30]},
        ).commit()
    )

    encap = tuple(
        context.ndb.routes.dump()
        .filter(oif=context.ndb.interfaces[ifname]['index'])
        .filter(lambda x: x.encap is not None)
        .select('encap')
        .transform(encap=lambda x: json.loads(x))
    )[0].encap

    assert isinstance(encap, list)
    assert encap[0]['label'] == 20
    assert encap[0]['bos'] == 0
    assert encap[1]['label'] == 30
    assert encap[1]['bos'] == 1


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_json(context):
    data = json.loads(''.join(context.ndb.interfaces.summary().format('json')))
    assert isinstance(data, list)
    for row in data:
        assert isinstance(row, dict)


class MD(csv.Dialect):
    quotechar = "'"
    doublequote = False
    quoting = csv.QUOTE_MINIMAL
    delimiter = ","
    lineterminator = "\n"


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_csv(context):
    record_length = 0

    for record in context.ndb.routes.dump():
        if record_length == 0:
            record_length = len(record)
        else:
            assert len(record) == record_length

    reader = csv.reader(context.ndb.routes.dump().format('csv'), dialect=MD())
    for record in reader:
        assert len(record) == record_length


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_nested_ipaddr(context):
    ifname = context.new_ifname
    ipaddr1 = context.new_ipaddr
    ipaddr2 = context.new_ipaddr

    (
        context.ndb.interfaces.create(ifname=ifname, kind='dummy', state='up')
        .add_ip(address=ipaddr1, prefixlen=24)
        .add_ip(address=ipaddr2, prefixlen=24)
        .commit()
    )

    records = repr(
        context.ndb.interfaces[ifname]
        .ipaddr.dump()
        .filter(lambda x: x.family == AF_INET)
    )
    rlen = len(records.split('\n'))
    # 2 ipaddr
    assert rlen == 2


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_nested_ports(context):
    ifbr0 = context.new_ifname
    ifbr0p0 = context.new_ifname
    ifbr0p1 = context.new_ifname

    with context.ndb.interfaces as i:
        i.create(ifname=ifbr0p0, kind='dummy').commit()
        i.create(ifname=ifbr0p1, kind='dummy').commit()
        (
            i.create(ifname=ifbr0, kind='bridge')
            .add_port(ifbr0p0)
            .add_port(ifbr0p1)
            .commit()
        )

    records = len(
        repr(context.ndb.interfaces[ifbr0].ports.summary()).split('\n')
    )
    # 1 port
    assert records == 2



import errno

import pytest
from pr2test.context_manager import make_test_matrix
from pr2test.marks import require_root
from pr2test.tools import qdisc_exists

from pyroute2 import NetlinkError

pytestmark = [require_root()]
test_matrix = make_test_matrix(targets=['local', 'netns'])


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_pfifo(context):
    index, ifname = context.default_interface
    context.ipr.tc('add', 'pfifo', index=index, limit=700)
    assert qdisc_exists(context.netns, 'pfifo', ifname=ifname, limit=700)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_pfifo_fast(context):
    index, ifname = context.default_interface
    context.ipr.tc('add', 'pfifo_fast', index=index, handle=0)
    ret = qdisc_exists(context.netns, 'pfifo_fast', ifname=ifname)[0]
    assert ret.get_attr('TCA_OPTIONS')['priomap']


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_plug(context):
    index, ifname = context.default_interface
    context.ipr.tc('add', 'plug', index=index, limit=13107)
    assert qdisc_exists(context.netns, 'plug', ifname=ifname)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_blackhole(context):
    index, ifname = context.default_interface
    context.ipr.tc('add', 'blackhole', index=index)
    assert qdisc_exists(context.netns, 'blackhole', ifname=ifname)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_codel(context):
    index, ifname = context.default_interface
    context.ipr.tc(
        'add',
        'codel',
        index=index,
        handle='1:0',
        cdl_interval='40ms',
        cdl_target='2ms',
        cdl_limit=5000,
        cdl_ecn=1,
    )
    assert qdisc_exists(
        context.netns, 'codel', ifname=ifname, codel_ecn=1, codel_limit=5000
    )


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_sfq(context):
    index, ifname = context.default_interface
    context.ipr.tc('add', 'sfq', index=index, handle=0, perturb=10)
    assert qdisc_exists(context.netns, 'sfq', ifname=ifname, perturb_period=10)


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_tbf(context):
    index, ifname = context.default_interface
    context.ipr.tc(
        'add',
        'tbf',
        index=index,
        handle=0,
        rate='220kbit',
        latency='50ms',
        burst=1540,
    )
    opts = qdisc_exists(context.netns, 'tbf', ifname=ifname)[0].get_nested(
        'TCA_OPTIONS', 'TCA_TBF_PARMS'
    )
    assert opts['rate'] == 27500


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_choke(context):
    index, ifname = context.default_interface
    try:
        context.ipr.tc(
            'add', 'choke', index=index, limit=5500, bandwith=3000, ecn=True
        )
    except NetlinkError as e:
        if e.code == errno.ENOENT:
            pytest.skip('qdisc not supported: choke')
        raise
    opts = qdisc_exists(context.netns, 'choke', ifname=ifname)[0].get_nested(
        'TCA_OPTIONS', 'TCA_CHOKE_PARMS'
    )
    assert opts['limit'] == 5500
    assert opts['qth_max'] == 1375
    assert opts['qth_min'] == 458



import errno

import pytest
from pr2test.context_manager import make_test_matrix
from pr2test.marks import require_root
from pr2test.tools import qdisc_exists

from pyroute2 import NetlinkError

pytestmark = [require_root()]
test_matrix = make_test_matrix(targets=['local', 'netns'])


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_drr(context):
    index, ifname = context.default_interface
    try:
        context.ipr.tc('add', 'drr', index=index, handle='1:')
    except NetlinkError as e:
        if e.code == errno.ENOENT:
            pytest.skip('qdisc not supported: drr')
        raise
    context.ipr.tc('add-class', 'drr', index=index, handle='1:20', quantum=20)
    context.ipr.tc('add-class', 'drr', index=index, handle='1:30', quantum=30)
    assert qdisc_exists(context.netns, 'drr', ifname=ifname)
    cls = context.ipr.get_classes(index=index)
    assert len(cls) == 2
    assert cls[0].get_attr('TCA_KIND') == 'drr'
    assert cls[1].get_attr('TCA_KIND') == 'drr'
    assert cls[0].get_attr('TCA_OPTIONS').get_attr('TCA_DRR_QUANTUM') == 20
    assert cls[1].get_attr('TCA_OPTIONS').get_attr('TCA_DRR_QUANTUM') == 30


@pytest.mark.parametrize('context', test_matrix, indirect=True)
def test_hfsc(context):
    index, ifname = context.default_interface
    # root queue
    context.ipr.tc('add', 'hfsc', index=index, handle='1:0', default='1:1')
    assert qdisc_exists(context.netns, 'hfsc', ifname=ifname, defcls=1)
    # classes
    context.ipr.tc(
        'add-class',
        'hfsc',
        index=index,
        handle='1:1',
        parent='1:0',
        rsc={'m2': '3mbit'},
    )
    cls = context.ipr.get_classes(index=index)
    assert len(cls) == 2  # implicit root class + the defined one
    assert cls[0].get_attr('TCA_KIND') == 'hfsc'
    assert cls[1].get_attr('TCA_KIND') == 'hfsc'
    curve = cls[1].get_attr('TCA_OPTIONS').get_attr('TCA_HFSC_RSC')
    assert curve['m1'] == 0
    assert curve['d'] == 0
    assert curve['m2'] == 375000
    assert cls[1].get_attr('TCA_OPTIONS').get_attr('TCA_HFSC_FSC') is None
    assert cls[1].get_attr('TCA_OPTIONS').get_attr('TCA_HFSC_USC') is None



import socket
import subprocess
import threading
import time

import pytest
from pr2test.marks import require_root

from pyroute2 import Conntrack, NFCTSocket, config
from pyroute2.netlink.nfnetlink.nfctsocket import NFCTAttrTuple

pytestmark = [
    pytest.mark.skipif(
        int(config.uname[2][0]) < 5,
        reason='skip conntrack tests on kernels < 5 for the time being',
    ),
    require_root(),
]


def server(address, port, env):
    ss = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    ss.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    ss.bind((address, port))
    ss.listen(1)
    conn, cadd = ss.accept()
    env['client'] = cadd
    conn.recv(16)
    conn.shutdown(socket.SHUT_RDWR)
    conn.close()
    ss.close()


class Client:
    def __init__(self, address, port):
        self.ss = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.ss.connect((address, port))

    def stop(self):
        self.ss.send(b'\x00' * 16)
        self.ss.shutdown(socket.SHUT_RDWR)
        self.ss.close()


class BasicSetup:
    def __init__(self, request, tmpdir):
        # run server / client
        self.env = {}
        self.ct = Conntrack()
        self.nfct = NFCTSocket()
        if self.ct.count() == 0:
            self.ct.close()
            self.nfct.close()
            pytest.skip('conntrack modules are not supported')
        self.server = threading.Thread(
            target=server, args=('127.0.0.1', 5591, self.env)
        )
        self.server.start()
        e = None
        for x in range(5):
            try:
                self.client = Client('127.0.0.1', 5591)
                time.sleep(1)
                break
            except Exception as exc:
                e = exc
        else:
            raise e

    def teardown(self):
        self.nfct.close()
        self.ct.close()
        self.client.stop()
        self.server.join()
        self.env = {}


class CheckEntries:
    def add_tuple(self, saddr, daddr, proto, sport, dport):
        tuple_orig = NFCTAttrTuple(
            saddr=saddr, daddr=daddr, proto=proto, sport=sport, dport=dport
        )
        self.tuples.append(tuple_orig)
        return tuple_orig

    def __init__(self, request, tmpdir):
        self.tuples = []
        self.COUNT_CT = 20
        self.ct = Conntrack()

        for sport in range(20000, 20000 + self.COUNT_CT):
            tuple_orig = self.add_tuple(
                saddr='192.168.122.1',
                daddr='192.168.122.67',
                proto=socket.IPPROTO_TCP,
                sport=sport,
                dport=5599,
            )
            self.ct.entry(
                'add',
                timeout=60,
                tuple_orig=tuple_orig,
                tuple_reply=tuple_orig.reverse(),
            )

    def teardown(self):
        for tuple_orig in self.tuples:
            self.ct.entry('del', tuple_orig=tuple_orig)
        self.ct.close()


@pytest.fixture
def ct_basic(request, tmpdir):
    ctx = BasicSetup(request, tmpdir)
    yield ctx
    ctx.teardown()


@pytest.fixture
def ct_inject(request, tmpdir):
    ctx = CheckEntries(request, tmpdir)
    yield ctx
    ctx.teardown()


def test_stat(ct_basic):
    stat = ct_basic.ct.stat()
    cpus = [
        x
        for x in (
            subprocess.check_output('cat /proc/cpuinfo', shell=True).split(
                b'\n'
            )
        )
        if x.startswith(b'processor')
    ]
    assert len(stat) == len(cpus)


def test_count_dump(ct_basic):
    # These values should be pretty the same, but the call is not atomic
    # so some sessions may end or begin that time.
    assert len(list(ct_basic.ct.dump())) > 0
    assert ct_basic.ct.count() > 0


def test_nfct_dump(ct_basic):
    # "grep" our client / server connection from the dump
    for connection in ct_basic.nfct.dump():
        addr = (
            connection.get_attr('CTA_TUPLE_ORIG')
            .get_attr('CTA_TUPLE_IP')
            .get_attr('CTA_IP_V4_SRC')
        )
        port = (
            connection.get_attr('CTA_TUPLE_ORIG')
            .get_attr('CTA_TUPLE_PROTO')
            .get_attr('CTA_PROTO_SRC_PORT')
        )
        if ct_basic.env['client'] == (addr, port):
            break
    else:
        raise Exception('connection not found')


def test_ct_dump(ct_inject):
    tuple_match = NFCTAttrTuple(saddr='192.168.122.1', daddr='192.168.122.67')

    count_found = 0
    tuple_filter = tuple_match
    for entry in ct_inject.ct.dump_entries(tuple_orig=tuple_match):
        count_found += 1
    assert count_found == ct_inject.COUNT_CT

    count_found = 0
    tuple_filter = NFCTAttrTuple(proto=socket.IPPROTO_TCP)
    for entry in ct_inject.ct.dump_entries(tuple_orig=tuple_filter):
        if tuple_match == entry.tuple_orig:
            count_found += 1

    assert count_found == ct_inject.COUNT_CT



import sqlite3
import sys
import uuid

import pytest


@pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason='SQLite3 backup not supported on this Python version',
)
def test_file_backup(context):
    filename = str(uuid.uuid4()) + '-backup.db'
    context.ndb.backup(filename)
    backup = sqlite3.connect(filename)
    cursor = backup.cursor()
    cursor.execute('SELECT f_IFLA_IFNAME FROM interfaces WHERE f_index > 0')
    interfaces_from_backup = {x[0] for x in cursor.fetchall()}
    interfaces_from_ndb = {
        x.ifname for x in context.ndb.interfaces.summary().select('ifname')
    }
    assert interfaces_from_ndb == interfaces_from_backup



def _test_ntables(self):
    setA = set(
        filter(
            lambda x: x is not None,
            [
                x.get_attr('NDTA_PARMS').get_attr('NDTPA_IFINDEX')
                for x in self.ip.get_ntables()
            ],
        )
    )
    setB = set([x['index'] for x in self.ip.get_links()])
    assert setA == setB
